import logging
import re
from copy import deepcopy
from typing import cast

from pydantic import ValidationError

from labels.model.file import DependencyType, Location, LocationReadCloser, Scope
from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.model.package import Digest, Language, Package, PackageType
from labels.model.relationship import Relationship, RelationshipType
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.javascript.model import PnpmEntry
from labels.parsers.cataloger.javascript.package import package_url
from labels.parsers.collection.yaml import parse_yaml_with_tree_sitter
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)

VERSION_PATTERN = re.compile(r"(\d+\.\d+\.\d+(-[0-9A-Za-z\.]+)?)")


def extract_package_name_from_key_dependency(item: str) -> str | None:
    # Regex pattern to extract the package name
    pattern = r"^@?[\w-]+/[\w-]+$"
    match = re.match(pattern, item)
    if match:
        return match.group(0)
    return None


def extract_version_from_value_dependency(item: str) -> str | None:
    # Regex pattern to extract the version number before any parentheses
    pattern = r"^(\d+\.\d+\.\d+)"
    match = re.match(pattern, item)
    if match:
        return match.group(1)
    return None


def _get_package(
    packages: list[Package],
    dep_name: str | None,
    dep_version: str | None,
) -> Package | None:
    return next(
        (x for x in packages if x.name == dep_name and x.version == dep_version),
        None,
    )


def _process_relationships(
    dependencies: IndexedDict[str, ParsedValue],
    packages: list[Package],
    current_package: Package | None,
) -> list[Relationship]:
    relationships: list[Relationship] = []
    for raw_dep_name, raw_dep_version in dependencies.items():
        if not isinstance(raw_dep_version, str):
            continue
        dep_name = extract_package_name_from_key_dependency(
            raw_dep_name,
        )
        dep_version = extract_version_from_value_dependency(
            raw_dep_version,
        )
        if (dep := _get_package(packages, dep_name, dep_version)) and current_package:
            relationships.append(
                Relationship(
                    from_=dep,
                    to_=current_package,
                    type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                ),
            )
    return relationships


def _generate_relations_relationship(
    package_yaml: IndexedDict[str, ParsedValue],
    packages: list[Package],
) -> list[Relationship]:
    relationships: list[Relationship] = []
    packages_items = package_yaml["packages"]
    if not isinstance(packages_items, IndexedDict):
        return relationships
    for package_key, package_value in packages_items.items():
        if not isinstance(package_value, IndexedDict):
            continue
        if match_ := re.search(r"/(@?[^@]+)@(\d+\.\d+\.\d+)", package_key):
            package_name = match_.groups()[0]
            package_version = match_.groups()[1]
            current_package = _get_package(
                packages,
                dep_name=package_name,
                dep_version=package_version,
            )
            dependencies = package_value.get("dependencies")
            if dependencies and isinstance(dependencies, IndexedDict):
                relationships.extend(
                    _process_relationships(dependencies, packages, current_package),
                )
    return relationships


def _get_package_metadata(
    package_value: IndexedDict[str, ParsedValue],
    *,
    is_dev: bool,
) -> PnpmEntry:
    resolution_value = package_value.get("resolution")
    integrity_value = (
        resolution_value.get("integrity") if isinstance(resolution_value, IndexedDict) else None
    )

    integrity_value = integrity_value if isinstance(integrity_value, str) else None

    return PnpmEntry(
        is_dev=is_dev if isinstance(is_dev, bool) else False,
        integrity=Digest(
            algorithm="sha-512",
            value=integrity_value,
        ),
    )


def manage_coordinates(
    package_yaml: IndexedDict[str, ParsedValue],
    package_key: str,
    package_name: str,
    direct_dependencies: list[ParsedValue],
    base_location: Location,
) -> Location:
    current_location = deepcopy(base_location)
    if current_location.coordinates:
        packages = package_yaml["packages"]
        if not isinstance(packages, IndexedDict):
            return current_location
        position = packages.get_key_position(package_key)
        current_location.coordinates.line = position.start.line
        is_transitive = package_name not in direct_dependencies
        current_location.dependency_type = (
            DependencyType.TRANSITIVE if is_transitive else DependencyType.DIRECT
        )
    return current_location


def process_package_string(
    package: str,
    spec: IndexedDict[str, ParsedValue],
) -> tuple[str, str] | None:
    if package.startswith("github"):
        pkg_name = spec.get("name", "")
        pkg_version = spec.get("version", "")
    else:
        pkg_info: list[str] = VERSION_PATTERN.split(package.strip("\"'"))
        if len(pkg_info) < 2:
            return None
        pkg_name = pkg_info[0].lstrip("/")[0:-1]
        pkg_version = pkg_info[1]

    if not isinstance(pkg_name, str) or not isinstance(pkg_version, str):
        return None
    return pkg_name, pkg_version


def _process_package(
    package_key: str,
    pkg_spec: IndexedDict[str, ParsedValue],
    package_yaml: IndexedDict[str, ParsedValue],
    direct_dependencies: list[ParsedValue],
    base_location: Location,
) -> Package | None:
    if match_ := process_package_string(package_key, pkg_spec):
        package_name = match_[0]
        package_version = match_[1]

        if not package_name or not package_version:
            return None
        is_dev = pkg_spec.get("dev")
        current_location: Location = manage_coordinates(
            package_yaml,
            package_key,
            package_name,
            direct_dependencies,
            base_location,
        )
        is_dev = is_dev if isinstance(is_dev, bool) else False
        current_location.scope = Scope.DEV if is_dev else Scope.PROD
        try:
            return Package(
                name=package_name,
                version=package_version,
                locations=[current_location],
                language=Language.JAVASCRIPT,
                licenses=[],
                type=PackageType.NpmPkg,
                p_url=package_url(package_name, package_version),
                metadata=_get_package_metadata(pkg_spec, is_dev=is_dev),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": current_location.path(),
                    },
                },
            )
    return None


def parse_pnpm_lock(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_yaml: IndexedDict[str, ParsedValue] = cast(
        IndexedDict[str, ParsedValue],
        parse_yaml_with_tree_sitter(reader.read_closer.read()),
    )

    if not package_yaml:
        return [], []
    dependencies = package_yaml.get("dependencies")
    dev_dependencies = package_yaml.get("devDependencies")
    direct_dependencies: list[ParsedValue] = []
    if isinstance(dependencies, IndexedList) and isinstance(dev_dependencies, IndexedList):
        direct_dependencies = [*dev_dependencies, *dependencies]
    packages: list[Package] = []
    relationships: list[Relationship] = []
    packages_items = package_yaml["packages"]
    if not isinstance(packages_items, IndexedDict):
        return [], []
    for package_key, pkg_spec in packages_items.items():
        if not isinstance(pkg_spec, IndexedDict):
            continue
        if package := _process_package(
            package_key,
            pkg_spec,
            package_yaml,
            direct_dependencies,
            reader.location,
        ):
            packages.append(package)

    relationships = _generate_relations_relationship(package_yaml, packages)

    return packages, relationships



from labels.enrichers.ruby.get import RubyGemsPackage, get_gem_package
from labels.enrichers.utils import infer_algorithm
from labels.model.package import Artifact, Digest, HealthMetadata, Package
from labels.utils.licenses.validation import validate_licenses


def _get_artifact(current_package: RubyGemsPackage) -> Artifact | None:
    digest_value = current_package.get("sha") or None
    return Artifact(
        url=current_package["gem_uri"],
        integrity=Digest(
            algorithm=infer_algorithm(digest_value),
            value=digest_value,
        ),
    )


def _set_health_metadata(
    package: Package,
    gem_package: RubyGemsPackage,
    current_package: RubyGemsPackage | None,
) -> None:
    package.health_metadata = HealthMetadata(
        latest_version=gem_package["version"],
        latest_version_created_at=gem_package["version_created_at"],
        authors=gem_package["authors"],
        artifact=_get_artifact(current_package) if current_package else None,
    )


def complete_package(package: Package) -> Package:
    current_package = get_gem_package(package.name, package.version)
    gem_package = get_gem_package(package.name)
    if not gem_package:
        return package

    _set_health_metadata(package, gem_package, current_package)

    gem_licenses = gem_package["licenses"]
    if gem_licenses:
        package.licenses = validate_licenses(gem_licenses)
    return package



import logging
from copy import deepcopy
from typing import cast

from labels.model.file import LocationReadCloser
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Package
from labels.model.relationship import Relationship, RelationshipType
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.javascript.package import new_package_lock_v1, new_package_lock_v2
from labels.parsers.collection.json import parse_json_with_tree_sitter

LOGGER = logging.getLogger(__name__)
EMPTY_DICT: IndexedDict[str, ParsedValue] = IndexedDict()


def get_name_from_path(name: str) -> str:
    return name.split("node_modules/")[-1]


def get_direct_dependencies(package_lock_path: IndexedDict[str, ParsedValue]) -> list[str]:
    all_dependencies: ParsedValue = package_lock_path.get("dependencies", IndexedDict())

    transitives: set[str] = set()
    for details in cast("IndexedDict[str, ParsedValue]", all_dependencies).values():
        if not isinstance(details, IndexedDict):
            continue
        reqs = cast("IndexedDict[str, ParsedValue]", details.get("requires", EMPTY_DICT))
        for dep in reqs:
            transitives.add(dep)

    return [
        dep
        for dep in cast("IndexedDict[str, ParsedValue]", all_dependencies)
        if dep not in transitives
    ]


def get_direct_dependencies_v2_v3(package_lock_path: IndexedDict[str, ParsedValue]) -> list[str]:
    all_dependencies: ParsedValue = package_lock_path.get("packages", IndexedDict())
    if not isinstance(all_dependencies, IndexedDict):
        LOGGER.warning("No direct deps found found in package JSON")
        return []

    result: list[str] = []
    for dep, value in all_dependencies.items():
        if isinstance(value, IndexedDict) and dep == "":
            deps_candidate: ParsedValue = value.get("dependencies", EMPTY_DICT)
            if isinstance(deps_candidate, IndexedDict):
                result.extend(deps_candidate)
            dev_deps_candidate: ParsedValue = value.get("devDependencies", EMPTY_DICT)
            if isinstance(dev_deps_candidate, IndexedDict):
                result.extend(dev_deps_candidate)
    return result


def _solve_sub_dependencies(
    reader: LocationReadCloser,
    sub_deps: IndexedDict[str, ParsedValue],
) -> list[Package]:
    packages = []
    for dep_key, dep_value in sub_deps.items():
        if isinstance(dep_value, IndexedDict):
            pkg = new_package_lock_v1(
                reader.location,
                dep_key,
                dep_value,
                is_transitive=True,
            )
            if pkg:
                packages.append(pkg)

            nested_sub_deps = dep_value.get("dependencies", EMPTY_DICT)
            if isinstance(nested_sub_deps, IndexedDict):
                packages.extend(_solve_sub_dependencies(reader, nested_sub_deps))
    return packages


def handle_v1(
    reader: LocationReadCloser,
    package_json: IndexedDict[str, ParsedValue],
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    relationships: list[Relationship] = []

    deps: ParsedValue = package_json.get("dependencies", IndexedDict())
    direct_dependencies = get_direct_dependencies(package_json)
    if not isinstance(deps, IndexedDict):
        LOGGER.warning("No packages found in package JSON")
        return ([], [])
    for dependency_key, dependency_value in deps.items():
        if not isinstance(dependency_value, IndexedDict):
            continue
        name: str = dependency_key
        is_transitive = name not in direct_dependencies
        if pkg := new_package_lock_v1(
            reader.location,
            name,
            dependency_value,
            is_transitive=is_transitive,
        ):
            packages.append(pkg)

        requires = [
            package
            for package in packages
            if package.name
            in cast("IndexedDict[str, ParsedValue]", dependency_value.get("requires", EMPTY_DICT))
        ]
        sub_deps: ParsedValue = dependency_value.get("dependencies", EMPTY_DICT)
        if sub_deps and isinstance(sub_deps, IndexedDict):
            packages.extend(_solve_sub_dependencies(reader, sub_deps))
        current_package: Package | None = next(
            (package for package in packages if package.name == dependency_key),
            None,
        )
        if not current_package:
            continue
        relationships.extend(
            [
                Relationship(
                    from_=require,
                    to_=current_package,
                    type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                )
                for require in requires
            ],
        )
    return packages, relationships


def _get_name(dependency_key: str, package_value: IndexedDict[str, ParsedValue]) -> str | None:
    name = dependency_key
    if not name:
        if "name" not in package_value:
            return None
        name = str(package_value["name"])

    # handle alias name
    if "name" in package_value and package_value["name"] != dependency_key:
        name = str(package_value["name"])

    return get_name_from_path(name)


def handle_v2(
    reader: LocationReadCloser,
    package_json: IndexedDict[str, ParsedValue],
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    relationships: list[Relationship] = []
    pkgs: ParsedValue = package_json.get("packages", IndexedDict())
    if not isinstance(pkgs, IndexedDict):
        LOGGER.warning("No packages found in package JSON")
        return ([], [])
    dependency_map: dict[str, ParsedValue] = {}
    direct_dependencies = get_direct_dependencies_v2_v3(package_json)
    for dependency_key, package_value in pkgs.items():
        if not dependency_key or not isinstance(package_value, IndexedDict):
            continue
        name = _get_name(dependency_key, package_value)
        is_transitive = name not in direct_dependencies
        if pkg := new_package_lock_v2(
            deepcopy(reader.location),
            get_name_from_path(name or dependency_key),
            package_value,
            is_transitive=is_transitive,
        ):
            packages.append(pkg)
            dependencies = package_value.get("dependencies", EMPTY_DICT)
            dependency_map[name or dependency_key] = dependencies

    for pkg in packages:
        dependencies = dependency_map.get(pkg.name, IndexedDict())
        if not isinstance(dependencies, IndexedDict):
            continue
        for dep_name in dependencies:
            dependency_pkg = next((p for p in packages if p.name == dep_name), None)
            if dependency_pkg:
                relationships.append(
                    Relationship(
                        from_=dependency_pkg,
                        to_=pkg,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                    ),
                )

    return packages, relationships


def parse_package_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_json: IndexedDict[str, ParsedValue] = cast(
        "IndexedDict[str, ParsedValue]",
        parse_json_with_tree_sitter(reader.read_closer.read()),
    )
    packages: list[Package] = []
    relationships: list[Relationship] = []

    match package_json.get("lockfileVersion", 1):
        case 1:
            return handle_v1(reader, package_json)
        case 2 | 3:
            return handle_v2(reader, package_json)
    return packages, relationships



import logging
from pathlib import Path
from typing import Unpack

import confuse

from labels.model.core import (
    AwsCredentialsConfig,
    LoadedConfig,
    OutputFormat,
    SbomConfig,
    ScanArgs,
    SourceType,
)
from labels.model.sources import AwsCredentials
from labels.utils.exceptions import InvalidConfigFileError

LOGGER = logging.getLogger(__name__)

YAML_SCHEMA = {
    "source": confuse.String(),
    "source_type": confuse.OneOf(
        [_source.value for _source in SourceType],
    ),
    "include": confuse.Optional(confuse.Sequence(confuse.String()), default=(".",)),
    "exclude": confuse.Optional(confuse.Sequence(confuse.String()), default=()),
    "execution_id": confuse.Optional(confuse.String(), default=None),
    "docker_credentials": confuse.Optional(
        {
            "username": confuse.String(),
            "password": confuse.String(),
        },
        default={},
    ),
    "aws_credentials": confuse.Optional(
        {
            "external_id": confuse.Optional(confuse.String(), default=None),
            "role": confuse.Optional(confuse.String(), default=None),
            "access_key_id": confuse.Optional(confuse.String(), default=None),
            "secret_access_key": confuse.Optional(confuse.String(), default=None),
            "session_token": confuse.Optional(confuse.String(), default=None),
        },
        default={},
    ),
    "output": {
        "name": confuse.String(),
        "format": confuse.OneOf(
            [_format.value for _format in OutputFormat],
        ),
    },
    "debug": confuse.Optional(confuse.OneOf([True, False]), default=False),
    "include_package_metadata": confuse.Optional(confuse.OneOf([True, False]), default=True),
}

STATIC_SCAN_YAML_SCHEMA = {
    "source_type": confuse.Optional(
        confuse.OneOf([_source.value for _source in SourceType]),
        default=SourceType.DIRECTORY.value,
    ),
    "working_dir": confuse.Optional(confuse.String(), default="./"),
    "execution_id": confuse.Optional(confuse.String(), default=None),
    "docker_credentials": confuse.Optional(
        {
            "username": confuse.String(),
            "password": confuse.String(),
        },
        default={},
    ),
    "aws_credentials": confuse.Optional(
        {
            "external_id": confuse.Optional(confuse.String(), default=None),
            "role": confuse.Optional(confuse.String(), default=None),
            "access_key_id": confuse.Optional(confuse.String(), default=None),
            "secret_access_key": confuse.Optional(confuse.String(), default=None),
            "session_token": confuse.Optional(confuse.String(), default=None),
        },
        default={},
    ),
    "sbom": {
        "include_package_metadata": confuse.Optional(confuse.OneOf([True, False]), default=True),
        "include": confuse.Optional(confuse.Sequence(confuse.String()), default=(".",)),
        "exclude": confuse.Optional(confuse.Sequence(confuse.String()), default=()),
        "output": {
            "name": confuse.String(),
            "format": confuse.OneOf(
                [_format.value for _format in OutputFormat],
            ),
        },
    },
}


def build_labels_config_from_args(arg: str, **kwargs: Unpack[ScanArgs]) -> SbomConfig:
    return SbomConfig(
        source=arg,
        source_type=SourceType.from_string(kwargs["source"]),
        execution_id=None,
        output_format=OutputFormat.from_string(kwargs["format"]),
        output=kwargs["output"],
        docker_user=kwargs["docker_user"],
        docker_password=kwargs["docker_password"],
        aws_external_id=kwargs["aws_external_id"],
        aws_role=kwargs["aws_role"],
        debug=kwargs["debug"],
    )


def build_labels_config_from_file(
    config_file_path: str,
    *,
    static_scan: bool = False,
) -> SbomConfig:
    if Path(config_file_path).is_file():
        if not config_file_path.endswith((".yaml", ".yml")):
            error_msg = "The configuration file must be a YAML format"
            raise InvalidConfigFileError(config_file_path, error_msg)
        if static_scan:
            return load_config_from_static_scan(config_file_path)
        return load_config_from_file(config_file_path)

    raise InvalidConfigFileError(config_file_path)


def load_aws_credentials(config_aws: AwsCredentialsConfig | None) -> AwsCredentials | None:
    if (
        config_aws
        and (access_key_id := config_aws.get("access_key_id")) is not None
        and (secret_access_key := config_aws.get("secret_access_key")) is not None
    ):
        return AwsCredentials(
            access_key_id=access_key_id,
            secret_access_key=secret_access_key,
            session_token=config_aws.get("session_token"),
        )

    return None


def load_config_from_file(config_path: str) -> SbomConfig:
    template = _build_template(config_path)
    check_unused_keys(template, set(YAML_SCHEMA.keys()))

    config_as_dict: LoadedConfig = template.get(YAML_SCHEMA)

    config_docker = config_as_dict["docker_credentials"]
    config_aws = config_as_dict["aws_credentials"]
    output = config_as_dict["output"]

    return SbomConfig(
        source=config_as_dict["source"],
        source_type=SourceType.from_string(config_as_dict["source_type"]),
        output_format=OutputFormat.from_string(output["format"]),
        output=output["name"],
        include=config_as_dict["include"],
        exclude=config_as_dict["exclude"],
        docker_user=config_docker["username"] if config_docker else None,
        docker_password=config_docker["password"] if config_docker else None,
        aws_external_id=config_aws.get("external_id") if config_aws else None,
        aws_role=config_aws.get("role") if config_aws else None,
        aws_credentials=load_aws_credentials(config_aws),
        execution_id=config_as_dict["execution_id"],
        debug=config_as_dict["debug"],
        include_package_metadata=config_as_dict["include_package_metadata"],
    )


def _build_template(config_path: str) -> confuse.Configuration:
    template = confuse.Configuration("labels", read=False)
    template.set_file(config_path)
    template.read(user=False, defaults=False)
    return template


def check_unused_keys(config: confuse.Configuration, expected_keys: set[str]) -> None:
    config_keys = set(config.keys())
    unused_keys = config_keys - expected_keys
    if unused_keys:
        unrecognized_keys = ", ".join(unused_keys)
        msg = (
            f"Some keys were not recognized: {unrecognized_keys}."
            " The analysis will be performed only using the supported keys"
            " and defaults."
        )
        LOGGER.warning(msg)


def load_config_from_static_scan(config_path: str) -> SbomConfig:
    template = _build_template(config_path)
    config_as_dict = template.get(STATIC_SCAN_YAML_SCHEMA)
    labels_config = config_as_dict["sbom"]
    output = labels_config["output"]
    config_docker = config_as_dict["docker_credentials"]
    config_aws = config_as_dict["aws_credentials"]

    return SbomConfig(
        source=config_as_dict["working_dir"],
        source_type=SourceType.from_string(config_as_dict["source_type"]),
        output_format=OutputFormat.from_string(output["format"]),
        output=output["name"],
        include=labels_config["include"],
        exclude=labels_config["exclude"],
        docker_user=config_docker["username"] if config_docker else None,
        docker_password=config_docker["password"] if config_docker else None,
        aws_external_id=config_aws["external_id"] if config_aws else None,
        aws_role=config_aws["role"] if config_aws else None,
        aws_credentials=load_aws_credentials(config_aws),
        execution_id=config_as_dict["execution_id"],
        debug=config_as_dict.get("debug", False),
        include_package_metadata=labels_config.get("include_package_metadata", True),
    )



import re

# Feel free to add support for additional licenses, but ensure that each key
# follows the SPDX standard identifier for consistency. Each entry in this
# dictionary should use the SPDX license ID as the key, with the value being a
# tuple containing:
#   - A regex pattern to match the license text
#   - A sanitized version of the license name
# This setup maintains alignment with the SPDX standard while allowing flexible
# pattern matching.

# Common words to be removed from license names to standardize them for the
# matching search thus reducing regex cases
name_sanitization_terms = ["the", "software", "version", "license", "licence"]
sanitization_pattern = re.compile(rf"\b(?:{'|'.join(name_sanitization_terms)})\b", re.IGNORECASE)

COMMON_LICENSES = {
    "0BSD": (
        r"\b0bsd\b",
        "0BSD",
    ),
    "Apache-2.0": (
        r"\bapache[-\s]?v?2(\.0)?\b",
        "Apache 2.0",
    ),
    "Apache-1.0": (
        r"\bapache[-\s]?v?1(\.0)?\b",
        "Apache 1.0",
    ),
    "Apache-1.1": (
        r"\bapache[-\s]?v?1\.1\b",
        "Apache 1.1",
    ),
    "BSD-2-Clause": (
        r"\bbsd[-\s]?2[-\s]?clause\b",
        "BSD 2-Clause",
    ),
    "BSD-3-Clause": (
        r"\bbsd[-\s]?3[-\s]?clause\b",
        "BSD 3-Clause",
    ),
    "CC-BY-4.0": (
        r"\bcc[-\s]?by[-\s]?4\.0\b",
        "Creative Commons Attribution 4.0 International",
    ),
    "CC-BY-SA-4.0": (
        r"\bcc[-\s]?by[-\s]?sa[-\s]?4\.0\b",
        "Creative Commons Attribution ShareAlike 4.0 International",
    ),
    "CC-BY-ND-4.0": (
        r"\bcc[-\s]?by[-\s]?nd[-\s]?4\.0\b",
        "Creative Commons Attribution No Derivatives 4.0 International",
    ),
    "CC-BY-NC-4.0": (
        r"\bcc[-\s]?by[-\s]?nc[-\s]?4\.0\b",
        "Creative Commons Attribution NonCommercial 4.0 International",
    ),
    "CC-BY-NC-SA-4.0": (
        r"\bcc[-\s]?by[-\s]?nc[-\s]?sa[-\s]?4\.0\b",
        "Creative Commons Attribution NonCommercial ShareAlike 4.0 International",
    ),
    "CC-BY-NC-ND-4.0": (
        r"\bcc[-\s]?by[-\s]?nc[-\s]?nd[-\s]?4\.0\b",
        "Creative Commons Attribution NonCommercial No Derivatives 4.0 International",
    ),
    "CC-BY-3.0": (
        r"\bcc[-\s]?by[-\s]?3\.0\b",
        "Creative Commons Attribution 3.0 Unported",
    ),
    "CC-BY-SA-3.0": (
        r"\bcc[-\s]?by[-\s]?sa[-\s]?3\.0\b",
        "Creative Commons Attribution ShareAlike 3.0 Unported",
    ),
    "CC-BY-ND-3.0": (
        r"\bcc[-\s]?by[-\s]?nd[-\s]?3\.0\b",
        "Creative Commons Attribution No Derivatives 3.0 Unported",
    ),
    "CC-BY-NC-3.0": (
        r"\bcc[-\s]?by[-\s]?nc[-\s]?3\.0\b",
        "Creative Commons Attribution NonCommercial 3.0 Unported",
    ),
    "CC-BY-NC-SA-3.0": (
        r"\bcc[-\s]?by[-\s]?nc[-\s]?sa[-\s]?3\.0\b",
        "Creative Commons Attribution NonCommercial ShareAlike 3.0 Unported",
    ),
    "CC-BY-NC-ND-3.0": (
        r"\bcc[-\s]?by[-\s]?nc[-\s]?nd[-\s]?3\.0\b",
        "Creative Commons Attribution NonCommercial No Derivatives 3.0 Unported",
    ),
    "EPL-2.0": (
        r"\bepl[-\s]?2(\.0)?\b|\beclipse public v2\.0\b",
        "Eclipse Public 2.0",
    ),
    "EPL-1.0": (
        r"\bepl[-\s]?1(\.0)?\b|\beclipse public v1\.0\b",
        "Eclipse Public 1.0",
    ),
    "GPL-2.0-only": (
        r"\bgpl[-\s]?2(\.0)[-\s]?only\b",
        "GNU General Public v2.0 only",
    ),
    "GPL-2.0-or-later": (
        r"\bgpl[-\s]?2(\.0)?[-\s]?(?:or[-\s]?later|\+)?\b",
        "GNU General Public v2.0 or later",
    ),
    "GPL-3.0-only": (
        r"\bgpl[-\s]?3(\.0)?\b|\bgpl[-\s]?v3(\.0)?\b",
        "GNU General Public v3.0 only",
    ),
    "GPL-3.0-or-later": (
        r"\b(?:gnu\s+)?gpl(?:v)?3(\.0)?[-\s]?(?:or[-\s]?later|\+)?\b",
        "GNU General Public v3.0 or later",
    ),
    "LGPL-2.1-only": (
        r"\blgpl[-\s]?2\.1[-\s]?only\b",
        "GNU Lesser General Public v2.1 only",
    ),
    "LGPL-2.1-or-later": (
        r"\blgpl[-\s]?2\.1?[-\s]?(?:or[-\s]?later|\+)?\b",
        "GNU Lesser General Public v2.1 or later",
    ),
    "LGPL-3.0-only": (
        r"\blgpl[-\s]?3(\.0)?\b|\blgpl[-\s]?v3(\.0)?\b|\blgpl[-\s]?v3\b",
        "GNU Lesser General Public v3.0 only",
    ),
    "LGPL-3.0-or-later": (
        r"\blgpl[-\s]?3(\.0)?[-\s]?(?:or[-\s]?later|\+)?\b|"
        r"\blgpl[-\s]?v3(\.0)?[-\s]?(?:or[-\s]?later|\+)?\b"
        r"\blgpl[-\s]?v3\+\b",
        "GNU Lesser General Public v3.0 or later",
    ),
    "ISC": (
        r"\bisc\b",
        "ISC",
    ),
    "MIT": (
        r"\bmit\b",
        "MIT",
    ),
    "MPL-2.0": (
        r"\bmpl[-\s]?2(\.0)\b",
        "Mozilla Public 2.0",
    ),
    "OSL-3.0": (
        r"\bosl[-\s]?3(\.0)\b",
        "Open 3.0",
    ),
    "WTFPL": (
        r"\bwtfpl\b",
        "Do What The F*** You Want To Public (WTFPL)",
    ),
    "Zlib": (
        r"\bzlib\b",
        "Zlib",
    ),
    "Artistic-2.0": (
        r"\bartistic[-\s]?2(\.0)?\b",
        "Artistic 2.0",
    ),
    "BSL-1.0": (
        r"\bboost[-\s]?1(\.0)?\b",
        "Boost 1.0",
    ),
    "BlueOak-1.0.0": (
        r"\bBlueOak[-\s]?1\.0\.0\b",
        "Blue Oak Model 1.0.0",
    ),
    "PSF-2.0": (
        r"\bPSF[-\s]?2(\.0)?\b",
        "Python Foundation 2.0",
    ),
    "CC0-1.0": (
        r"\bCC0[-\s]?1\.0\b",
        "CC0-1.0",
    ),
    "Python-2.0": (
        r"\bpython[-\s]?2\.0\b",
        "Python 2.0",
    ),
}

INVERSE_LICENSES = {full_name: key for key, (_, full_name) in COMMON_LICENSES.items()}



import logging
import threading
from datetime import UTC, datetime

from requests import exceptions
from urllib3.exceptions import MaxRetryError

from labels.enrichers.java.get import MavenPackageInfo, get_maven_package_info, search_maven_package
from labels.enrichers.utils import infer_algorithm
from labels.model.package import Artifact, Digest, HealthMetadata, Package
from labels.parsers.cataloger.java.model import JavaArchive
from labels.parsers.cataloger.java.package import group_id_from_java_metadata
from labels.utils.licenses.validation import validate_licenses
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)
stop_event = threading.Event()


def parse_package_name(combined: str) -> tuple[str | None, str | None]:
    parts = combined.split(":")
    if len(parts) == 1:
        return None, parts[0]
    return parts[0], parts[1]


def _get_health_metadata_artifact(
    current_package: MavenPackageInfo | None,
) -> Artifact | None:
    if current_package and current_package.jar_url:
        digest_value = current_package.hash or None
        return Artifact(
            url=current_package.jar_url,
            integrity=Digest(
                algorithm=infer_algorithm(digest_value),
                value=digest_value,
            ),
        )
    return None


def get_group_id(package: Package) -> str | None:
    if isinstance(package.metadata, JavaArchive) and (
        g_id := group_id_from_java_metadata(package.name, package.metadata)
    ):
        return g_id
    if package_candidate := search_maven_package(package.name, package.version):
        return package_candidate.group
    return None


def _set_health_metadata(
    package: Package,
    maven_package: MavenPackageInfo,
    current_package: MavenPackageInfo | None,
) -> None:
    authors = maven_package.authors or []
    package.health_metadata = HealthMetadata(
        latest_version=maven_package.latest_version,
        latest_version_created_at=datetime.fromtimestamp(maven_package.release_date, tz=UTC)
        if maven_package.release_date
        else None,
        authors=", ".join(authors) if authors else None,
        artifact=_get_health_metadata_artifact(current_package),
    )


def _set_licenses(package: Package, maven_package: MavenPackageInfo) -> None:
    maven_package_licenses = maven_package.licenses or []
    if maven_package_licenses:
        package.licenses = validate_licenses(
            [package_license for package_license in maven_package_licenses if package_license],
        )


def get_complete_package(
    package: Package,
    group_id: str,
    artifact_id: str,
) -> MavenPackageInfo | None:
    maven_package = get_maven_package_info(group_id, artifact_id)
    if not maven_package:
        if package_candidate := search_maven_package(package.name, package.version):
            maven_package = get_maven_package_info(package_candidate.group, artifact_id)
        if not maven_package:
            return None
    return maven_package


def complete_package(package: Package) -> Package:
    if stop_event.is_set():
        return package

    group_id, artifact_id = parse_package_name(package.name)

    try:
        group_id = group_id or get_group_id(package)
        if not group_id or not artifact_id:
            return package

        maven_package = get_complete_package(package, group_id, artifact_id)
        if not maven_package:
            return package

        current_package = get_maven_package_info(group_id, artifact_id, package.version)

        _set_health_metadata(package, maven_package, current_package)
        _set_licenses(package, maven_package)

    except (exceptions.ConnectionError, TimeoutError, MaxRetryError, exceptions.RetryError) as ex:
        LOGGER.warning(
            "Failed to connect to the package manager.",
            extra={"extra": {"exception": format_exception(str(ex))}},
        )
        stop_event.set()
    return package



from datetime import datetime

import pydantic

from labels.enrichers.golang.get import fetch_latest_version_info, fetch_license_info
from labels.model.package import HealthMetadata, Package


class GolangModuleEntry(pydantic.BaseModel):
    h1_digest: str


def complete_package(package: Package) -> Package:
    latest = fetch_latest_version_info(package.name)
    if not latest:
        return package
    package.health_metadata = HealthMetadata(
        latest_version=latest["Version"],
        latest_version_created_at=datetime.fromisoformat(latest["Time"]),
        artifact=None,
    )
    if package.name.startswith("github.com"):
        license_info = fetch_license_info("/".join(package.name.split("/")[1:]))
        if not license_info:
            return package
        licenses = license_info["license"]["spdx_id"]
        package.licenses = [licenses]
    return package



import logging
from datetime import datetime
from typing import Any
from xml.etree import ElementTree as ET

from defusedxml import ElementTree
from pydantic import BaseModel, ConfigDict

from labels.enrichers.api_interface import make_get

LOGGER = logging.getLogger(__name__)


class MavenSearchDocResponse(BaseModel):
    id_: str
    group: str
    artifact: str
    version: str
    timestamp: int
    extra_classifiers: list[str]
    packaging: str | None = None
    tags: list[str] | None = None
    model_config = ConfigDict(frozen=True)


class AuthorsLicenses(BaseModel):
    authors: list[str]
    licenses: list[str]
    model_config = ConfigDict(frozen=True)


class MavenPackageInfo(BaseModel):
    group_id: str
    artifact_id: str
    latest_version: str | None = None
    release_date: int | None = None
    authors: list[str] | None = None
    licenses: list[str] | None = None
    version: str | None = None
    jar_url: str | None = None
    hash: str | None = None
    model_config = ConfigDict(frozen=True)


def build_maven_search_url(
    artifact_id: str,
    version: str | None,
) -> tuple[str, dict[str, str | int]]:
    base_url = "https://search.maven.org/solrsearch/select"
    query = f"a:{artifact_id} AND v:{version}" if version else f"a:{artifact_id}"
    params: dict[str, str | int] = {"q": query, "rows": 5, "wt": "json"}
    return base_url, params


def parse_maven_search_response(
    docs: list[dict[str, Any]],
    artifact_id: str,
    version: str,
) -> MavenSearchDocResponse | None:
    try:
        return MavenSearchDocResponse(
            id_=docs[0]["id"],
            group=docs[0]["g"],
            artifact=docs[0]["a"],
            version=docs[0]["v"] if "v" in docs[0] else docs[0]["latestVersion"],
            packaging=docs[0].get("p", None),
            timestamp=int(docs[0]["timestamp"] // 1000),
            extra_classifiers=docs[0]["ec"],
            tags=docs[0].get("tags", None),
        )
    except KeyError as exc:
        LOGGER.exception(
            "Error parsing Maven search response",
            extra={
                "extra": {
                    "artifact_id": artifact_id,
                    "version": version,
                    "doc": docs[0],
                    "key": exc.args[0],
                },
            },
        )
        return None


def search_maven_package(artifact_id: str, version: str) -> MavenSearchDocResponse | None:
    base_url, params = build_maven_search_url(artifact_id, version)
    package_data = make_get(base_url, params=params, timeout=30)
    if not package_data:
        return None

    docs = package_data["response"].get("docs", [])
    if not docs or len(docs) > 1:
        return None

    return parse_maven_search_response(docs, artifact_id, version)


def build_maven_urls(group_id: str, artifact_id: str, version: str) -> tuple[str, str, str]:
    group_id_path = group_id.replace(".", "/")
    base_path = f"https://repo1.maven.org/maven2/{group_id_path}/{artifact_id}/{version}"
    pom_url = f"{base_path}/{artifact_id}-{version}.pom"
    jar_url = f"{base_path}/{artifact_id}-{version}.jar"
    hash_url = f"{base_path}/{artifact_id}-{version}.jar.sha1"
    return pom_url, jar_url, hash_url


def get_package_hash(hash_url: str) -> str:
    hash_response: str | None = make_get(hash_url, content=True, timeout=30)
    return hash_response.strip() if hash_response else "Hash not available"


def parse_pom_xml(pom_content: str) -> ET.Element:
    return ElementTree.fromstring(pom_content)


def extract_authors_licenses(pom_xml: ET.Element) -> AuthorsLicenses:
    namespace = {"m": "http://maven.apache.org/POM/4.0.0"}
    authors = [
        author.text
        for author in pom_xml.findall("m:developers/m:developer/m:name", namespace)
        if author.text
    ]
    licenses = [
        getattr(package_license.find("m:name", namespace), "text", "")
        for package_license in pom_xml.findall("m:licenses/m:license", namespace)
    ]
    return AuthorsLicenses(authors=authors, licenses=licenses)


def get_authors_licenses(group_id: str, artifact_id: str, version: str) -> AuthorsLicenses | None:
    pom_url, _, _ = build_maven_urls(group_id, artifact_id, version)
    response = make_get(pom_url, content=True, timeout=30)
    if not response:
        return None

    pom_xml = parse_pom_xml(response)
    return extract_authors_licenses(pom_xml)


def parse_metadata_xml(metadata_content: str) -> tuple[str | None, int | None]:
    metadata_xml = ElementTree.fromstring(metadata_content)
    latest_stable_version = metadata_xml.find("versioning/release")
    release_date_tag = metadata_xml.find("versioning/lastUpdated")

    release_date: int | None = None
    if release_date_tag is not None and release_date_tag.text:
        release_date = int(
            datetime.strptime(release_date_tag.text + "+0000", "%Y%m%d%H%M%S%z").timestamp(),
        )

    return (
        latest_stable_version.text if latest_stable_version is not None else None,
        release_date,
    )


def build_metadata_url(group_id: str, artifact_id: str) -> str:
    group_id_path = group_id.replace(".", "/")
    return f"https://repo1.maven.org/maven2/{group_id_path}/{artifact_id}/maven-metadata.xml"


def get_maven_metadata(group_id: str, artifact_id: str) -> tuple[str | None, int | None]:
    metadata_url = build_metadata_url(group_id, artifact_id)
    response = make_get(metadata_url, content=True, timeout=30)
    if not response:
        return None, None

    return parse_metadata_xml(response)


def get_specific_version_info(
    group_id: str,
    artifact_id: str,
    version: str,
) -> MavenPackageInfo | None:
    pom_url, jar_url, hash_url = build_maven_urls(group_id, artifact_id, version)
    response = make_get(pom_url, timeout=30, content=True)
    if not response:
        return None

    package_hash = get_package_hash(hash_url)

    return MavenPackageInfo(
        group_id=group_id,
        artifact_id=artifact_id,
        version=version,
        jar_url=jar_url,
        hash=package_hash,
    )


def get_latest_version_info(group_id: str, artifact_id: str) -> MavenPackageInfo | None:
    latest_version, release_date = get_maven_metadata(group_id, artifact_id)
    if not latest_version:
        return None

    authors_licenses = get_authors_licenses(group_id, artifact_id, latest_version)

    return MavenPackageInfo(
        group_id=group_id,
        artifact_id=artifact_id,
        latest_version=latest_version,
        release_date=release_date,
        authors=authors_licenses.authors if authors_licenses else None,
        licenses=authors_licenses.licenses if authors_licenses else None,
    )


def get_maven_package_info(
    group_id: str,
    artifact_id: str,
    version: str | None = None,
) -> MavenPackageInfo | None:
    if version:
        return get_specific_version_info(group_id, artifact_id, version)
    return get_latest_version_info(group_id, artifact_id)



import os
import re
from typing import TypedDict, cast

from pydantic import BaseModel, ConfigDict

from labels.model.file import LocationReadCloser
from labels.model.package import Digest, Package
from labels.model.relationship import Relationship, RelationshipType
from labels.model.resolver import Resolver
from labels.parsers.cataloger.alpine.package import (
    ApkDBEntry,
    ApkFileRecord,
    ParsedData,
    new_package,
)
from labels.parsers.cataloger.generic.parser import Environment

APK_DB_GLOB = "**/lib/apk/db/installed"


class FileInfo(TypedDict):
    uid: str
    gid: str
    perms: str


class ApkField(BaseModel):
    name: str
    value: str
    model_config = ConfigDict(frozen=True)


class ApkFileParsingContext(BaseModel):
    files: list[ApkFileRecord]
    index_of_latest_directory: int
    index_of_latest_regular_file: int


def _parse_list_values(value: str | None, delimiter: str | None = None) -> list[str]:
    delimiter = delimiter or " "
    if not value:
        return []
    return value.split(delimiter)


def _process_file_info(
    info: str,
) -> FileInfo | None:
    file_info = info.split(":")
    if len(file_info) < 3:
        return None
    return {"uid": file_info[0], "gid": file_info[1], "perms": file_info[2]}


def process_checksum(value: str) -> Digest:
    algorithm = "md5"
    if value.startswith("Q1"):
        algorithm = "'Q1'+base64(sha1)"
    return Digest(algorithm=algorithm, value=value)


def parse_package(package: str) -> ParsedData | None:
    data: dict[str, str] = {}
    ctx = ApkFileParsingContext(
        files=[],
        index_of_latest_directory=-1,
        index_of_latest_regular_file=-1,
    )
    lines = package.split("\n")
    key = ""
    for line in lines:
        key = process_line(line, key, data, ctx)
    return construct_apk(data, ctx)


def process_line(line: str, key: str, data: dict[str, str], ctx: ApkFileParsingContext) -> str:
    if ":" in line:
        key, value = line.split(":", 1)
        data[key] = value
        update_context_with_line(key, value, ctx)
    elif key and key in data:
        data[key] += "\n" + line.strip()
    return key


def update_context_with_line(key: str, value: str, ctx: ApkFileParsingContext) -> None:
    match key:
        case "F":
            ctx.files.append(ApkFileRecord(path=os.path.join("/", value)))
            ctx.index_of_latest_directory = len(ctx.files) - 1
        case "M":
            index = ctx.index_of_latest_directory
            latest = ctx.files[index]
            if file_info := _process_file_info(value):
                latest.owner_uid = file_info["uid"]
                latest.owner_gid = file_info["gid"]
                latest.permissions = file_info["perms"]
            ctx.files[index] = latest
        case "R":
            index = ctx.index_of_latest_directory
            if index < 0:
                regular_file = os.path.join("/", value)
            else:
                latest_dir_path = ctx.files[index].path
                regular_file = os.path.join(latest_dir_path, value)
            ctx.files.append(ApkFileRecord(path=regular_file))
            ctx.index_of_latest_regular_file = len(ctx.files) - 1
        case "a":
            index = ctx.index_of_latest_regular_file
            latest = ctx.files[index]
            if file_info := _process_file_info(value):
                latest.owner_uid = file_info["uid"]
                latest.owner_gid = file_info["gid"]
                latest.permissions = file_info["perms"]
            ctx.files[index] = latest
        case "Z":
            index = ctx.index_of_latest_regular_file
            latest = ctx.files[index]
            latest.digest = process_checksum(value)


def construct_apk(data: dict[str, str], ctx: ApkFileParsingContext) -> ParsedData | None:
    if not (package := data.get("P")) or not (version := data.get("V")):
        return None

    return ParsedData(
        apk_db_entry=ApkDBEntry(
            package=package,
            origin_package=data.get("o"),
            maintainer=data.get("m"),
            version=version,
            architecture=data.get("A"),
            url=data.get("U", ""),
            description=data.get("T", ""),
            size=data.get("S", ""),
            installed_size=data.get("I"),
            dependencies=_parse_list_values(data.get("D")),
            provides=_parse_list_values(data.get("p")),
            checksum=data.get("C"),
            git_commit=data.get("c"),
            files=ctx.files,
        ),
        license=data.get("L"),
    )


def _build_lookup_table(pkgs: list[Package]) -> dict[str, list[Package]]:
    lookup: dict[str, list[Package]] = {}

    for pkg in pkgs:
        apkg: ApkDBEntry = cast(ApkDBEntry, pkg.metadata)
        if pkg.name not in lookup:
            lookup[pkg.name] = [pkg]
        else:
            lookup[pkg.name].append(pkg)

        for provides in apkg.provides:
            provides_k = strip_version_specifier(provides)
            if provides_k not in lookup:
                lookup[provides_k] = [pkg]
            else:
                lookup[provides_k].append(pkg)

    return lookup


def discover_package_dependencies(
    pkgs: list[Package],
) -> list[Relationship]:
    lookup: dict[str, list[Package]] = _build_lookup_table(pkgs)
    relationships: list[Relationship] = []

    for pkg in pkgs:
        apkg = cast(ApkDBEntry, pkg.metadata)
        for dep_specifier in apkg.dependencies:
            dep = strip_version_specifier(dep_specifier)
            relationships.extend(
                Relationship(
                    from_=dep_pk,
                    to_=pkg,
                    type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                )
                for dep_pk in lookup.get(dep, [])
            )
    return relationships


def strip_version_specifier(version: str) -> str:
    splitted_version: list[str] = re.split("[<>=]", version)
    return splitted_version[0]


def parse_apk_db(
    _resolver: Resolver,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]] | None:
    content = reader.read_closer.read()
    apks = [
        parsed_package
        for package in content.strip().split("\n\n")
        if package and (parsed_package := parse_package(package)) is not None
    ]

    entries = [
        entry
        for apk in apks
        if (entry := new_package(apk, _env.linux_release if _env else None, reader.location))
        is not None
    ]

    return entries, discover_package_dependencies(entries)



from typing import TypedDict, cast

import requests

from labels.config.cache import dual_cache


class NugetCatalogEntry(TypedDict, total=False):
    authors: str
    created: str
    description: str
    id: str
    isPrerelease: bool
    lastEdited: str
    licenseExpression: str
    listed: bool
    packageHash: str
    packageHashAlgorithm: str
    packageSize: int
    projectUrl: str
    published: str
    requireLicenseAcceptance: bool
    title: str
    verbatimVersion: str
    version: str
    tags: list[str]


def _fetch_nuget_data(url: str) -> dict | None:
    response = requests.get(url, timeout=30)
    if response.status_code != 200:
        return None
    return response.json()


def _get_catalog_entry_from_items(items_list: list) -> dict | None:
    try:
        return next(x["catalogEntry"] for x in reversed(items_list) if "catalogEntry" in x)
    except StopIteration:
        try:
            return next(
                y["catalogEntry"]
                for x in reversed(items_list)
                for y in reversed(x.get("items", []))
                if "catalogEntry" in y and "pre" not in y["catalogEntry"].get("version", "")
            )
        except StopIteration:
            return None


def _get_package_data_for_version(package_name: str, version: str) -> dict | None:
    base_url = f"https://api.nuget.org/v3/registration5-gz-semver2/{package_name}/{version}.json"
    package_data = _fetch_nuget_data(base_url)
    if not package_data:
        return None

    catalog_url = package_data.get("catalogEntry")
    if not isinstance(catalog_url, str):
        return None

    return _fetch_nuget_data(catalog_url)


def _get_package_index(package_name: str) -> dict | None:
    base_url = f"https://api.nuget.org/v3/registration5-gz-semver2/{package_name}/index.json"
    return _fetch_nuget_data(base_url)


def _get_latest_version_group_url(package_index: dict) -> str | None:
    items = package_index.get("items")
    if not isinstance(items, list) or not items:
        return None

    last_item = items[-1]
    return last_item.get("@id") if isinstance(last_item.get("@id"), str) else None


def _get_version_group_items(items_url: str) -> list:
    items_response = _fetch_nuget_data(items_url)
    if not items_response:
        return []
    return items_response.get("items", [])


def _get_latest_package_data(package_name: str) -> dict | None:
    package_index = _get_package_index(package_name)
    if not package_index:
        return None

    items_url = _get_latest_version_group_url(package_index)
    if not items_url:
        return None

    items_list = _get_version_group_items(items_url)
    return _get_catalog_entry_from_items(items_list)


@dual_cache
def get_nuget_package(package_name: str, version: str | None = None) -> NugetCatalogEntry | None:
    package_name = package_name.lower()

    if version:
        package_data = _get_package_data_for_version(package_name, version)
    else:
        package_data = _get_latest_package_data(package_name)

    if package_data is None:
        return None

    return cast(NugetCatalogEntry, package_data)



import hashlib
import logging
import os
import shutil
import tempfile
from collections.abc import Callable
from io import BufferedReader
from pathlib import Path
from typing import TextIO, cast
from zipfile import ZipInfo

from pydantic import BaseModel, ConfigDict, ValidationError

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Digest, Language, Package
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.java.archive_filename import ArchiveFilename, parse_filename
from labels.parsers.cataloger.java.model import (
    JavaArchive,
    JavaManifest,
    JavaPomProject,
    JavaPomProperties,
)
from labels.parsers.cataloger.java.package import group_id_from_java_metadata, package_url
from labels.parsers.cataloger.java.parse_java_manifest import (
    parse_java_manifest,
    select_licenses,
    select_name,
    select_version,
)
from labels.parsers.cataloger.java.parse_pom_properties import parse_pom_properties
from labels.parsers.cataloger.java.parse_pom_xml import ParsedPomProject, parse_pom_xml_project
from labels.utils.licenses import parser
from labels.utils.strings import format_exception
from labels.utils.zip import contents_from_zip, new_zip_file_manifest, zip_glob_match

LOGGER = logging.getLogger(__name__)


def new_digests_from_file(file_object: BufferedReader, hashes: list[str]) -> list[Digest]:
    hash_objects = [hashlib.new(hash_key, usedforsecurity=False) for hash_key in hashes]

    while chunk := file_object.read(4096):
        for hasher in hash_objects:
            hasher.update(chunk)

    if file_object.tell() == 0:
        return []

    return [
        Digest(
            algorithm=hash_name,
            value=hasher.hexdigest(),
        )
        for hash_name, hasher in zip(hashes, hash_objects, strict=False)
    ]


def get_digests_from_archive(archive_path: str) -> list[Digest]:
    with Path(archive_path).open("rb") as reader:
        return new_digests_from_file(reader, ["sha1"])


def pom_properties_by_parent(
    archive_path: str,
    extract_paths: list[str],
) -> dict[str, JavaPomProperties]:
    properties_by_parent_path = {}
    contents_of_maven_properties = contents_from_zip(archive_path, *extract_paths)

    for file_path, file_contents in contents_of_maven_properties.items():
        if not file_contents:
            continue
        pom_properties = parse_pom_properties(file_path, file_contents)
        if not pom_properties:
            continue
        if not pom_properties.group_id or not pom_properties.version:
            continue
        properties_by_parent_path[str(Path(file_path).parent)] = pom_properties

    return properties_by_parent_path


def pom_project_by_parent(
    archive_path: str,
    location: Location,
    extract_paths: list[str],
) -> dict[str, ParsedPomProject]:
    contents_of_maven_project = contents_from_zip(archive_path, *extract_paths)

    project_by_parent = {}

    for file_path, file_contents in contents_of_maven_project.items():
        pom_project = parse_pom_xml_project(file_path, file_contents, location)
        if not pom_project:
            continue
        if (
            not pom_project.java_pom_project.parent and not pom_project.java_pom_project.version
        ) or not pom_project.java_pom_project.artifact_id:
            continue

        project_by_parent[str(Path(file_path).parent)] = pom_project

    return project_by_parent


def artifact_id_matches_filename(artifact_id: str, filename: str) -> bool:
    if not artifact_id or not filename:
        return False

    return artifact_id.startswith(filename) or filename.startswith(artifact_id)


class ArchiveParser(BaseModel):
    file_manifest: list[ZipInfo]
    location: Location
    archive_path: str | None
    content_path: str
    file_info: ArchiveFilename
    detect_nested: bool
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def parse(self) -> tuple[list[Package] | None, list[Relationship] | None]:
        packages = []
        parent_pkg = self.discover_main_package()
        if not parent_pkg:
            return None, None
        aux_pkgs = self.discover_pkgs_from_all_maven_files(parent_pkg)

        packages = [parent_pkg, *aux_pkgs]

        return packages, []

    def discover_pkgs_from_all_maven_files(self, parent_pkg: Package | None) -> list[Package]:
        if not parent_pkg or not self.archive_path:
            return []

        pkgs: list[Package] = []

        properties = pom_properties_by_parent(
            self.archive_path,
            zip_glob_match(self.file_manifest, case_sensitive=False, patterns=("*pom.properties",)),
        )

        projects = pom_project_by_parent(
            self.archive_path,
            self.location,
            zip_glob_match(self.file_manifest, case_sensitive=False, patterns=("*pom.xml",)),
        )

        for parent_path, properties_obj in properties.items():
            pom_project: ParsedPomProject | None = projects.get(parent_path, None)
            if pkg_from_pom := new_package_from_maven_data(
                properties_obj,
                pom_project,
                parent_pkg,
                self.location,
            ):
                pkgs.append(pkg_from_pom)
        return pkgs

    def discover_main_package(self) -> Package | None:
        manifest_matches = zip_glob_match(
            self.file_manifest,
            case_sensitive=False,
            patterns=("/META-INF/MANIFEST.MF",),
        )
        if not manifest_matches or not self.archive_path:
            return None
        contents = contents_from_zip(self.archive_path, *manifest_matches)
        if not contents:
            return None

        manifest_contents = contents[manifest_matches[0]]
        manifest = parse_java_manifest(manifest_contents)
        licenses_, name, version = self.parse_licenses(manifest)
        metadata = JavaArchive(
            virtual_path=self.location.path(),
            manifest=manifest,
            archive_digests=get_digests_from_archive(self.archive_path),
        )

        if not name or not version:
            return None

        try:
            return Package(
                name=name,
                version=version,
                licenses=licenses_,
                locations=[self.location],
                type=self.file_info.pkg_type(),
                language=Language.JAVA,
                metadata=metadata,
                p_url=package_url(name, version, metadata),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": self.location.path(),
                    },
                },
            )
            return None

    def get_license_from_file_in_archive(self) -> list[str]:
        file_licenses = []
        for filename in parser.LICENSES_FILE_NAMES:
            license_matches = zip_glob_match(
                self.file_manifest,
                case_sensitive=True,
                patterns=(f"/META-INF/{filename}",),
            )
            if not license_matches:
                license_matches = zip_glob_match(
                    self.file_manifest,
                    case_sensitive=True,
                    patterns=(f"/{filename}",),
                )

            if license_matches and self.archive_path:
                contents = contents_from_zip(self.archive_path, *license_matches)
                for license_match in license_matches:
                    license_contents = contents.get(license_match, "")

                    parsed = parser.parse_license(license_contents)
                    if parsed:
                        file_licenses.extend(parsed)

        return file_licenses

    def parse_licenses(self, manifest: JavaManifest) -> tuple[list[str], str, str]:
        licenses_ = select_licenses(manifest)

        (
            name,
            version,
            pom_licenses,
        ) = self.guess_main_package_name_and_version_from_pom()

        if not name:
            name = select_name(manifest, self.file_info)
        if not version:
            version = select_version(manifest, self.file_info)

        if not licenses_:
            licenses_.extend(pom_licenses or [])

        if not licenses_:
            file_licenses = self.get_license_from_file_in_archive()
            if file_licenses:
                licenses_.extend(file_licenses)
        return licenses_, name, version

    def extract_properties_and_projects(
        self,
    ) -> tuple[dict[str, JavaPomProperties], dict[str, ParsedPomProject]]:
        properties = {}
        projects = {}

        pom_property_matches = zip_glob_match(
            self.file_manifest,
            case_sensitive=False,
            patterns=("*pom.properties",),
        )
        pom_matches = zip_glob_match(
            self.file_manifest,
            case_sensitive=False,
            patterns=("*pom.xml",),
        )
        if self.archive_path:
            properties = pom_properties_by_parent(self.archive_path, pom_property_matches)
            projects = pom_project_by_parent(self.archive_path, self.location, pom_matches)

        return properties, projects

    def find_relevant_objects(
        self,
        properties: dict[str, JavaPomProperties],
        projects: dict[str, ParsedPomProject],
    ) -> tuple[JavaPomProperties | None, ParsedPomProject | None]:
        for parent_path, properties_obj in properties.items():
            if (
                properties_obj.artifact_id
                and artifact_id_matches_filename(properties_obj.artifact_id, self.file_info.name)
                and (proj := projects.get(parent_path))
            ):
                return properties_obj, proj
        return None, None

    def extract_name_version(
        self,
        properties_obj: JavaPomProperties,
        project_obj: ParsedPomProject,
    ) -> tuple[str | None, str | None]:
        name = properties_obj.artifact_id if properties_obj else None
        version = properties_obj.version if properties_obj else None

        if not name and project_obj:
            name = project_obj.java_pom_project.artifact_id
        if not version and project_obj:
            version = project_obj.java_pom_project.version

        return name, version

    def guess_main_package_name_and_version_from_pom(
        self,
    ) -> tuple[str | None, str | None, list[str] | None]:
        properties, projects = self.extract_properties_and_projects()
        properties_obj, project_obj = self.find_relevant_objects(properties, projects)

        if not properties_obj or not project_obj:
            return None, None, None

        name, version = self.extract_name_version(properties_obj, project_obj)

        return name, version, []


def save_archive_to_tmp(
    archive_virtual_path: str,
    _reader: TextIO,
) -> tuple[str | None, str | None, Callable[[], None]]:
    name = Path(archive_virtual_path).name
    temp_dir = tempfile.mkdtemp(prefix="sbom-archive-contents-")

    def cleanup_fn() -> None:
        shutil.rmtree(temp_dir)

    content_dir = os.path.join(temp_dir, "contents")
    Path(content_dir).mkdir()

    archive_path = os.path.join(temp_dir, f"archive-{name}")
    shutil.copy(archive_virtual_path, archive_path)

    return content_dir, archive_path, cleanup_fn


def new_java_archive_parser(
    reader: LocationReadCloser,
    *,
    detect_nested: bool,
) -> tuple[ArchiveParser | None, Callable[[], None] | None]:
    if not reader.location.coordinates:
        return None, None
    current_file_path = reader.location.coordinates.real_path
    content_path, archive_path, cleanup_fn = save_archive_to_tmp(
        current_file_path,
        reader.read_closer,
    )
    if not archive_path or not content_path:
        logging.error("unable to read files from java archive")
        return None, None

    file_info = parse_filename(current_file_path)
    file_manifest = new_zip_file_manifest(archive_path)
    return (
        ArchiveParser(
            file_manifest=file_manifest,
            location=reader.location,
            archive_path=archive_path,
            content_path=content_path,
            file_info=file_info,
            detect_nested=detect_nested,
        ),
        cleanup_fn,
    )


def parse_java_archive(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    parser, _file_cleanup_fn = new_java_archive_parser(reader, detect_nested=True)
    if parser is None:
        return [], []

    pkgs, relations = parser.parse()
    return pkgs or [], relations or []


def new_package_from_maven_data(
    pom_properties: JavaPomProperties,
    _parsed_pom_project: ParsedPomProject | None,
    parent_pkg: Package,
    location: Location,
) -> Package | None:
    name = pom_properties.artifact_id
    version = pom_properties.version

    if not name or not version:
        return None

    v_path_suffix = ""

    parent_metadata: JavaArchive = cast(JavaArchive, parent_pkg.metadata)
    group_id = group_id_from_java_metadata(parent_pkg.name, parent_metadata)
    parent_key = f"{group_id}:{parent_pkg.name}:{parent_pkg.version}"
    pom_project_key = f"{pom_properties.group_id}:{name}:{version}"

    if parent_key != pom_project_key:
        v_path_suffix += f":{pom_properties.group_id}:{name}"

    virtual_path = f"{location.path()}{v_path_suffix}"
    pkg_pom_project: JavaPomProject | None = None

    metadata = JavaArchive(
        virtual_path=virtual_path,
        pom_properties=pom_properties,
        pom_project=pkg_pom_project,
        parent=parent_pkg,
    )

    try:
        return Package(
            name=name,
            version=version,
            licenses=[],
            locations=[location],
            type=pom_properties.pkg_type_indicated(),
            language=Language.JAVA,
            metadata=metadata,
            p_url=package_url(name, version, metadata),
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": location.path(),
                },
            },
        )
        return None



import gzip
import logging
import os
from contextlib import suppress
from datetime import UTC, datetime
from pathlib import Path
from typing import Any, TextIO, cast

from labels.model.file import LocationReadCloser
from labels.model.package import Digest, Package
from labels.model.relationship import Relationship
from labels.model.resolver import FileReader, Resolver
from labels.parsers.cataloger.arch.package import AlpmDBEntry, AlpmFileRecord, new_package
from labels.parsers.cataloger.generic.parser import Environment

LOGGER = logging.getLogger(__name__)

IGNORED_FILES = {
    "/set": True,
    "/.BUILDINFO": True,
    "/.PKGINFO": True,
    "": True,
}


def parse_pkg_files(pkg_fields: dict[str, Any]) -> AlpmDBEntry | None:
    entry = AlpmDBEntry()

    entry.licenses = pkg_fields.get("license", "")
    entry.base_package = pkg_fields.get("base", "")
    entry.package = pkg_fields.get("name", "")
    entry.version = pkg_fields.get("version", "")
    entry.description = pkg_fields.get("desc", "")
    entry.architecture = pkg_fields.get("arch", "")
    entry.size = pkg_fields.get("size", 0)
    entry.packager = pkg_fields.get("packager", "")
    entry.url = pkg_fields.get("url", "")
    entry.validation = pkg_fields.get("validation", "")
    entry.reason = pkg_fields.get("reason", 0)
    entry.files = [
        AlpmFileRecord(path=item["path"], digests=item.get("digests"))
        for item in pkg_fields.get("files", [])
    ]
    entry.backup = [
        AlpmFileRecord(path=item["path"], digests=item["digests"])
        for item in pkg_fields.get("backup", [])
    ]

    if not entry.package and not entry.files and not entry.backup:
        return None
    return entry


def parse_key_value_pair(line: str) -> dict[str, Any]:
    try:
        key, value = line.split("\n", 1)
    except ValueError:
        return {}
    key = key.replace("%", "").lower()
    value = value.strip()

    if key == "files":
        return {key: _parse_files(value)}
    if key == "backup":
        return {key: parse_backup(value)}
    if key in ["reason", "size"]:
        try:
            return {key: parse_numeric_field(key, value)}
        except ValueError:
            LOGGER.exception("Failed to parse %s to integer: %s", key, value)
            return {}

    return {key: value}


def _parse_files(value: str) -> list[dict[str, Any]]:
    files = []
    for item in value.split("\n"):
        path = f"/{item}"
        if path not in IGNORED_FILES:
            files.append({"path": path})
    return files


def parse_backup(value: str) -> list[dict[str, Any]]:
    backup = []
    for item in value.split("\n"):
        backup_fields = item.split("\t", 1)
        path = f"/{backup_fields[0]}"
        if path not in IGNORED_FILES:
            backup.append(
                {
                    "path": path,
                    "digests": [Digest(algorithm="md5", value=backup_fields[1])],
                },
            )
    return backup


def parse_numeric_field(key: str, value: str) -> int:
    try:
        return int(value)
    except ValueError as exc:
        error_msg = f"Failed to parse {key} to integer: {value}"
        raise ValueError(error_msg) from exc


def parse_alpm_db_entry(reader: TextIO) -> AlpmDBEntry | None:
    pkg_fields: dict[str, Any] = {}
    lines = reader.read().split("\n\n")
    for line in lines:
        if not line.strip():
            break  # End of block or file
        pkg_fields.update(parse_key_value_pair(line))

    return parse_pkg_files(pkg_fields)


def parse_mtree(reader: TextIO) -> list[AlpmFileRecord]:
    file_info: dict[str, dict[str, str]] = {}
    result_records: list[AlpmFileRecord] = []
    data = reader.read()

    for line in data.splitlines():
        if line.startswith(".") and "time=" in line:
            parts = line.strip().split()
            file_path = parts[0][1:]
            file_attributes = {}
            for part in parts[1:]:
                if "=" in part:
                    key, value = part.split("=", 1)
                    file_attributes[key] = value
            file_info[file_path] = file_attributes
    for file_path, file_attributes in file_info.items():
        if file_path.startswith("/."):
            continue
        result = AlpmFileRecord(
            path=file_path,
            size=file_attributes.get("size"),
            time=datetime.fromtimestamp(float(file_attributes["time"]), tz=UTC),
            type=file_attributes.get("type"),
            digests=[
                Digest(
                    algorithm=key.replace("digest", ""),
                    value=value,
                )
                for key, value in file_attributes.items()
                if key.endswith("digest")
            ],
            uid=file_attributes.get("uid"),
            gid=file_attributes.get("gid"),
            link=file_attributes.get("link"),
        )
        result_records.append(result)
    return result_records


def parse_alpm_db(
    _resolver: Resolver,
    _env: Environment,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    data = parse_alpm_db_entry(reader.read_closer)
    if not data or not reader.location.coordinates:
        LOGGER.warning("No data or location found")
        return ([], [])
    with suppress(IndexError):
        mtree_location = _resolver.files_by_path(
            os.path.join(Path(reader.location.coordinates.real_path).parent, "mtree"),
        )[0]
        if mtree_reader := _resolver.file_contents_by_location(
            mtree_location,
            function_reader=cast(FileReader, gzip.open),
            mode="rt",
        ):
            data.files = parse_mtree(mtree_reader)
    with suppress(IndexError):
        files_location = _resolver.files_by_path(
            os.path.join(Path(reader.location.coordinates.real_path).parent, "files"),
        )[0]
        if (files_reader := _resolver.file_contents_by_location(files_location)) and (
            files_metadata := parse_alpm_db_entry(files_reader)
        ):
            data.backup = files_metadata.backup

    package = new_package(data, _env.linux_release, reader.location)

    return [package] if package else [], []



import logging
from typing import TYPE_CHECKING

from labels.utils.strings import format_exception

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import ItemsView

from typing import cast

from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.file import DependencyType, Location, LocationReadCloser, Scope
from labels.model.indexables import IndexedDict
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.collection.yaml import parse_yaml_with_tree_sitter

LOGGER = logging.getLogger(__name__)


def _get_location(location: Location, sourceline: int, *, is_dev: bool = False) -> Location:
    location.scope = Scope.DEV if is_dev else Scope.PROD
    if location.coordinates:
        c_upd = {"line": sourceline}
        l_upd = {"coordinates": location.coordinates.model_copy(update=c_upd)}
        location.dependency_type = DependencyType.DIRECT
        return location.model_copy(update=l_upd)
    return location


def _get_packages(
    reader: LocationReadCloser,
    dependencies: IndexedDict[str, str] | None,
    *,
    is_dev: bool = False,
) -> list[Package]:
    if dependencies is None:
        return []
    packages = []
    items: ItemsView[str, IndexedDict[str, str] | str] = dependencies.items()
    for name, version in items:
        if not name or not isinstance(version, str) or not version:
            continue
        general_location = _get_location(
            reader.location,
            dependencies.get_key_position(name).start.line,
            is_dev=is_dev,
        )
        try:
            packages.append(
                Package(
                    name=name,
                    version=version,
                    locations=[general_location],
                    language=Language.DART,
                    licenses=[],
                    type=PackageType.DartPubPkg,
                    p_url=PackageURL(  # type: ignore
                        type="pub",
                        name=name,
                        version=version,
                    ).to_string(),
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": general_location.path(),
                    },
                },
            )
            continue
    return packages


def parse_pubspec_yaml(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    content = cast(
        "IndexedDict[str, IndexedDict[str, str]]",
        parse_yaml_with_tree_sitter(reader.read_closer.read()),
    )

    if not content:
        return [], []

    deps: IndexedDict[str, str] | None = content.get("dependencies")
    dev_deps: IndexedDict[str, str] | None = content.get("dev_dependencies")
    packages = [
        *_get_packages(reader, deps, is_dev=False),
        *_get_packages(reader, dev_deps, is_dev=True),
    ]
    return packages, []



import re

from labels.model.file import (
    LocationReadCloser,
)
from labels.model.package import Package
from labels.model.relationship import (
    Relationship,
)
from labels.model.resolver import (
    Resolver,
)
from labels.parsers.cataloger.cpp.package import new_conan_file_dep
from labels.parsers.cataloger.generic.parser import (
    Environment,
)


def parse_conan_file(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    line_deps: bool = False
    is_dev = False
    for line_number, line in enumerate(
        reader.read_closer.read().splitlines(),
        1,
    ):
        if re.search(r"^\[(tool|build)_requires\]$", line):
            line_deps = True
            is_dev = True
        elif not is_dev and line.startswith("[requires]"):
            line_deps = True
            is_dev = False
        elif line_deps:
            if not line or line.startswith("["):
                line_deps = False
                continue
            new_conan_file_dep(
                line,
                reader,
                line_number,
                is_dev=is_dev,
                packages=packages,
            )
    return packages, []



import re
from copy import deepcopy

from packageurl import PackageURL

from labels.model.file import DependencyType, Location, LocationReadCloser, Scope
from labels.model.package import Language, Package, PackageType
from labels.parsers.cataloger.graph_parser import Graph, NId, node_to_str


def get_conan_dep_info(dep_line: str) -> tuple[str, str]:
    product, version = re.sub(r"[\"\]\[]", "", dep_line).strip().split("@")[0].split("/")
    if "," in version:
        version = re.sub(r",(?=[<>=])", " ", version).split(",")[0]
    return product, version


def build_conan_file_location(
    location: Location,
    line_number: int,
    *,
    is_dev: bool,
) -> Location:
    location = deepcopy(location)
    location.scope = Scope.DEV if is_dev else Scope.PROD
    if location.coordinates:
        location.coordinates.line = line_number
        location.dependency_type = DependencyType.DIRECT
    return location


def build_conan_file_py_location(location: Location, sourceline: int, *, is_dev: bool) -> Location:
    location.scope = Scope.DEV if is_dev else Scope.PROD
    if location.coordinates:
        c_upd = {"line": sourceline}
        l_upd = {"coordinates": location.coordinates.model_copy(update=c_upd)}
        return location.model_copy(update=l_upd)
    return location


def package_url(name: str, version: str) -> str:
    return PackageURL(
        type="conan",
        namespace="",
        name=name,
        version=version,
        qualifiers=None,
        subpath="",
    ).to_string()


def new_conan_file_py_dep(
    *,
    graph: Graph,
    node_id: NId,
    dep_info: str | None = None,
    is_dev: bool,
    location: Location,
) -> Package:
    dep_attrs = graph.nodes[node_id]
    if dep_info is None:
        dep_info = dep_attrs.get("label_text") or node_to_str(graph, node_id)
    dep_info = dep_info.replace("(", "").replace(")", "")
    product, version = get_conan_dep_info(dep_info)
    location.scope = Scope.DEV if is_dev else Scope.PROD
    dep_line = dep_attrs["label_l"]
    location = build_conan_file_py_location(location, dep_line, is_dev=is_dev)
    return Package(
        name=product,
        version=version,
        type=PackageType.ConanPkg,
        locations=[location],
        p_url=package_url(product, version),
        metadata=None,
        language=Language.CPP,
        licenses=[],
        is_dev=is_dev,
    )


def new_conan_file_dep(
    line: str,
    reader: LocationReadCloser,
    line_number: int,
    *,
    is_dev: bool,
    packages: list[Package],
) -> None:
    pkg_name, pkg_version = get_conan_dep_info(line)
    location = build_conan_file_location(reader.location, line_number, is_dev=is_dev)
    packages.append(
        Package(
            name=pkg_name,
            version=pkg_version,
            type=PackageType.ConanPkg,
            locations=[location],
            p_url=package_url(pkg_name, pkg_version),
            metadata=None,
            language=Language.CPP,
            licenses=[],
            is_dev=is_dev,
        ),
    )


def new_conan_lock_dep(
    dep_info: str,
    location: Location,
    *,
    is_dev: bool = False,
) -> Package:
    product, version = dep_info.split("/")
    version = version.split("#")[0]
    location.scope = Scope.DEV if is_dev else Scope.PROD
    return Package(
        name=product,
        version=version,
        type=PackageType.ConanPkg,
        locations=[location],
        p_url=package_url(product, version),
        metadata=None,
        language=Language.CPP,
        licenses=[],
        is_dev=is_dev,
    )



import logging
from datetime import datetime

from packageurl import PackageURL
from pydantic import BaseModel, ValidationError

from labels.model.file import Location
from labels.model.package import Digest, Language, Package, PackageType
from labels.model.release import Release
from labels.parsers.cataloger.utils import purl_qualifiers
from labels.utils.licenses.validation import validate_licenses
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


class AlpmFileRecord(BaseModel):
    path: str
    type: str | None = None
    uid: str | None = None
    gid: str | None = None
    time: datetime | None = None
    size: str | None = None
    link: str | None = None
    digests: list[Digest] | None = None


class AlpmDBEntry(BaseModel):
    licenses: str = ""
    base_package: str = ""
    package: str = ""
    version: str = ""
    description: str = ""
    architecture: str = ""
    size: int = 0
    packager: str = ""
    url: str = ""
    validation: str = ""
    reason: int = 0
    files: list[AlpmFileRecord] | None = None
    backup: list[AlpmFileRecord] | None = None


def package_url(entry: AlpmDBEntry, distro: Release | None = None) -> str:
    qualifiers = {"arch": entry.architecture}
    if entry.base_package:
        qualifiers["upstream"] = entry.base_package
    return PackageURL(
        type="alpm",
        name=entry.package,
        version=entry.version,
        qualifiers=purl_qualifiers(qualifiers, distro),  # type: ignore
        subpath="",
    ).to_string()


def new_package(
    entry: AlpmDBEntry,
    release: Release | None,
    db_location: Location,
) -> Package | None:
    name = entry.package
    version = entry.version

    if not name or not version:
        return None

    licenses_candidates = entry.licenses.split("\n")

    try:
        return Package(
            name=name,
            version=version,
            locations=[db_location],
            licenses=validate_licenses(licenses_candidates),
            type=PackageType.AlpmPkg,
            metadata=entry,
            p_url=package_url(entry, release),
            language=Language.UNKNOWN_LANGUAGE,
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": db_location.path(),
                },
            },
        )
        return None



import logging
import re
from copy import deepcopy

from pydantic import BaseModel, ValidationError

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.java.model import JavaArchive, JavaPomProject
from labels.parsers.cataloger.java.package import package_url
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)
# Constants
QUOTE = r'["\']'
NL = r"(\n?\s*)?"
TEXT = r'[^"\']+'
RE_LINE_COMMENT: re.Pattern[str] = re.compile(r"^.*" rf"{NL}//" r".*$")


def is_comment(line: str) -> bool:
    return (
        line.strip().startswith("//")
        or line.strip().startswith("/*")
        or line.strip().endswith("*/")
    )


class LockFileDependency(BaseModel):
    group: str
    name: str
    version: str
    line: int | None = None


def create_java_archive(
    dependency: LockFileDependency,
    name: str,
    version: str,
) -> JavaArchive:
    return JavaArchive(
        pom_project=JavaPomProject(
            group_id=dependency.group,
            name=name,
            artifact_id=name,
            version=version,
        ),
    )


def create_package(
    name: str,
    version: str,
    location: Location,
    archive: JavaArchive,
) -> Package:
    return Package(
        name=name,
        version=version,
        locations=[location],
        language=Language.JAVA,
        type=PackageType.JavaPkg,
        metadata=archive,
        p_url=package_url(name, version, archive),
        licenses=[],
    )


def avoid_cmt(line: str, *, is_block_cmt: bool) -> tuple[str, bool]:
    if RE_LINE_COMMENT.match(line):
        line = line.split("//", 1)[0]
    if is_block_cmt:
        if "*/" in line:
            is_block_cmt = False
            line = line.split("*/", 1).pop()
        else:
            return "", is_block_cmt
    if "/*" in line:
        line_cmt_open = line.split("/*", 1)[0]
        if "*/" in line:
            line = line_cmt_open + line.split("*/", 1).pop()
        else:
            line = line_cmt_open
            is_block_cmt = True
    return line, is_block_cmt


def get_line_number(content: str, match_start: int) -> int:
    return content[:match_start].count("\n") + 2


def extract_gradle_configs(content: str) -> set[str]:
    config_pattern = re.compile(r"configurations\s*\{([^}]+)\}", re.DOTALL)
    custom_config_pattern = re.compile(r"\s*(\w+)\s*")
    configs = {
        "runtimeOnly",
        "api",
        "compile",
        "compileOnly",
        "implementation",
    }

    config_blocks = config_pattern.findall(content)
    for block in config_blocks:
        custom_configs = custom_config_pattern.findall(block)
        configs.update(custom_configs)

    return configs


def build_regex_with_configs(configs: set[str]) -> dict[str, re.Pattern[str]]:
    config_pattern = "|".join(configs)

    return {
        "RE_GRADLE_A": re.compile(
            rf"^{NL}(?:{config_pattern}){NL}[(]?{NL}"
            rf"group{NL}:{NL}{QUOTE}(?P<group>{TEXT}){QUOTE}{NL},"
            rf"{NL}name{NL}:{NL}{QUOTE}(?P<name>{TEXT}){QUOTE}{NL}"
            rf"(?:,{NL}version{NL}:{NL}{QUOTE}(?P<version>{TEXT}){QUOTE}{NL})"
            rf"?.*$",
        ),
        "RE_GRADLE_B": re.compile(
            rf"^.*{NL}(?:{config_pattern}){NL}[(]?{NL}{QUOTE}(?P<statement>{TEXT}){QUOTE}",
        ),
        "RE_GRADLE_C": re.compile(
            rf"{NL}(?:{config_pattern}){NL}\("
            rf"{NL}{QUOTE}(?P<statement>{TEXT}){QUOTE}{NL}\)"
            rf"{NL}{{({NL})version{NL}{{({NL})strictly{NL}\({NL}"
            rf"{QUOTE}(?P<version>{TEXT}){QUOTE}{NL}\){NL}}}{NL}}}",
            re.DOTALL,
        ),
        "BLOCK": re.compile(
            rf"{NL}(?:{config_pattern}){NL}\("
            rf"{NL}{QUOTE}(?P<statement>{TEXT}){QUOTE}{NL}\)"
            rf"{NL}\{{(.*?version{NL}\{{.*?\}}){NL}\}}",
            re.DOTALL,
        ),
        "VERSION": re.compile(
            rf"version{NL}{{({NL})strictly{NL}\("
            rf"{NL}{QUOTE}(?P<version>{TEXT}){QUOTE}{NL}\){NL}}}",
            re.DOTALL,
        ),
    }


def get_block_deps(content: str, regexes: dict[str, re.Pattern[str]]) -> list[LockFileDependency]:
    dependencies = []

    for block in regexes["BLOCK"].finditer(content):
        product = block.group("statement")
        hit = regexes["VERSION"].search(block.group())
        if not hit:
            continue

        version = hit.group("version")
        if version == "":
            continue

        line_no = get_line_number(content, block.start())

        dependencies.append(
            LockFileDependency(
                group=product.split(":")[0],
                name=product,
                version=version,
                line=line_no,
            ),
        )
    return dependencies


def parse_dependencies(reader: LocationReadCloser) -> list[LockFileDependency]:
    content = reader.read_closer.read()
    configs = extract_gradle_configs(content)
    maven_regexes = build_regex_with_configs(configs)
    dependencies: list[LockFileDependency] = get_block_deps(content, maven_regexes)

    is_block_cmt = False
    for line_no, raw_line in enumerate(content.splitlines(), start=1):
        line, is_block_cmt = avoid_cmt(raw_line, is_block_cmt=is_block_cmt)

        if match := maven_regexes["RE_GRADLE_A"].match(line):
            group = match.group("group")
            product = group + ":" + match.group("name")
            version = match.group("version") or ""
        elif match := maven_regexes["RE_GRADLE_B"].match(line):
            statement = match.group("statement")
            product, version = (
                statement.rsplit(":", maxsplit=1) if statement.count(":") >= 2 else (statement, "")
            )
            group = product.split(":")[0]
        else:
            continue

        # Assuming a wildcard in Maven if the version is not found can
        # result in issues.
        # https://gitlab.com/fluidattacks/universe/-/issues/5635
        if version == "" or re.match(r"\${.*}", version):
            continue

        dependencies.append(
            LockFileDependency(
                group=group,
                name=product,
                version=version,
                line=line_no,
            ),
        )

    return dependencies


def create_packages(
    dependencies: list[LockFileDependency],
    reader_location: Location,
) -> list[Package]:
    packages: list[Package] = []

    for dependency in dependencies:
        name = dependency.name
        version = dependency.version

        if not name or not version:
            continue

        location = deepcopy(reader_location)
        if location.coordinates:
            location.coordinates.line = dependency.line

        archive = create_java_archive(dependency, name, version)

        try:
            package = create_package(name, version, location, archive)
            packages.append(package)
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": location.path(),
                    },
                },
            )

    return packages


def parse_gradle(
    _resolver: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    dependencies = parse_dependencies(reader)
    packages = create_packages(dependencies, reader.location)
    return packages, []



import logging
import os
from copy import deepcopy
from pathlib import Path
from typing import TextIO, cast

from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.file import Location
from labels.model.package import Language, Package, PackageType
from labels.model.release import Release
from labels.model.resolver import Resolver
from labels.parsers.cataloger.debian.model import DpkgDBEntry, DpkgFileRecord
from labels.parsers.cataloger.debian.parse_copyright import parse_licenses_from_copyright
from labels.parsers.cataloger.debian.parse_dpkg_info_files import (
    parse_dpkg_conffile_info,
    parse_dpkg_md5_info,
)
from labels.parsers.cataloger.utils import purl_qualifiers
from labels.utils.licenses.validation import validate_licenses
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def package_url(pkg: DpkgDBEntry, distro: Release | None = None) -> str:
    qualifiers = {"arch": pkg.architecture}
    if distro and (distro.id_ == "debian" or "debian" in (distro.id_like or [])):
        if distro.version_id:
            qualifiers["distro_version_id"] = distro.version_id
        qualifiers["distro_id"] = distro.id_
    if pkg.source:
        qualifiers["upstream"] = (
            f"{pkg.source}@{pkg.source_version}" if pkg.source_version else pkg.source
        )

    return PackageURL(
        type="deb",
        namespace=distro.id_ if distro and distro.id_ else "",
        name=pkg.package,
        version=pkg.version,
        qualifiers=purl_qualifiers(qualifiers, distro),
        subpath="",
    ).to_string()


def md5_key(metadata: DpkgDBEntry) -> str:
    content_key = metadata.package
    if metadata.architecture not in ("", "all"):
        return f"{content_key}:{metadata.architecture}"
    return content_key


def fetch_md5_content(
    resolver: Resolver,
    db_location: Location,
    entry: DpkgDBEntry,
) -> tuple[TextIO | None, Location | None] | None:
    if not db_location.coordinates:
        return None
    search_path = str(Path(db_location.coordinates.real_path).parent)
    if not search_path.endswith("status.d"):
        search_path = os.path.join(search_path, "info")
    name = md5_key(entry)
    location = resolver.relative_file_path(
        db_location,
        os.path.join(search_path, name + ".md5sums"),
    )
    if not location:
        location = resolver.relative_file_path(
            db_location,
            os.path.join(search_path, entry.package + ".md5sums"),
        )
    if not location:
        return None

    reader = resolver.file_contents_by_location(location)
    if not reader:
        LOGGER.warning(
            "failed to fetch deb md5 contents (package=%s)",
            entry.package,
        )
    return reader, location


def fetch_conffile_contents(
    resolver: Resolver,
    db_location: Location,
    entry: DpkgDBEntry,
) -> tuple[TextIO | None, Location | None] | None:
    if not db_location.coordinates:
        return None
    parent_path = str(Path(db_location.coordinates.real_path).parent)

    name = md5_key(entry)
    location = resolver.relative_file_path(
        db_location,
        os.path.join(parent_path, "info", name + ".conffiles"),
    )
    if not location:
        location = resolver.relative_file_path(
            db_location,
            os.path.join(parent_path, "info", entry.package + ".conffiles"),
        )
    if not location:
        return None, None
    reader = resolver.file_contents_by_location(location)
    if not reader:
        LOGGER.warning(
            "failed to fetch deb conffiles contents (package=%s)",
            entry.package,
        )
    return reader, location


def get_additional_file_listing(
    resolver: Resolver,
    db_location: Location,
    entry: DpkgDBEntry,
) -> tuple[list[DpkgFileRecord], list[Location]]:
    files: list[DpkgFileRecord] = []
    locations: list[Location] = []
    md5_result = fetch_md5_content(resolver, db_location, entry)
    if not md5_result:
        return files, locations
    md5_reader, md5_location = md5_result
    if md5_reader is not None and md5_location is not None:
        files.extend(parse_dpkg_md5_info(md5_reader))
        locations.append(md5_location)
    conffiles = fetch_conffile_contents(resolver, db_location, entry)
    if not conffiles:
        return files, locations
    conffiles_reader, conffiles_location = conffiles

    if conffiles_reader is not None and conffiles_location is not None:
        files.extend(parse_dpkg_conffile_info(conffiles_reader))
        locations.append(conffiles_location)

    return files, locations


def merge_file_listing(resolver: Resolver, db_location: Location, pkg: Package) -> None:
    metadata: DpkgDBEntry = cast(DpkgDBEntry, pkg.metadata)
    files, info_locations = get_additional_file_listing(resolver, db_location, metadata)

    if metadata.files is None:
        metadata.files = []

    for new_file in files:
        exists = False
        for existing_file in metadata.files:
            if existing_file.path == new_file.path:
                exists = True
                break
        if not exists:
            metadata.files.append(new_file)

    metadata.files = sorted(metadata.files, key=lambda x: x.path)
    pkg.metadata = metadata
    pkg.locations.extend(info_locations)


def fetch_copyright_contents(
    resolver: Resolver | None,
    db_location: Location,
    metadata: DpkgDBEntry,
) -> tuple[TextIO | None, Location | None]:
    if not resolver:
        return None, None

    copyright_path = os.path.join("/usr/share/doc", metadata.package, "copyright")
    location = resolver.relative_file_path(db_location, copyright_path)

    if not location:
        return None, None

    reader = resolver.file_contents_by_location(location)
    if not reader:
        LOGGER.warning(
            "Failed to fetch deb copyright contents (package=%s)",
            metadata.package,
        )

    return reader, location


def add_licenses(resolver: Resolver, db_location: Location, pkg: Package) -> None:
    metadata: DpkgDBEntry = cast(DpkgDBEntry, pkg.metadata)

    pkg.licenses = []
    copyright_reader, copyright_location = fetch_copyright_contents(resolver, db_location, metadata)

    if copyright_reader is not None and copyright_location is not None:
        licenses_strs = parse_licenses_from_copyright(copyright_reader)
        pkg.licenses = validate_licenses(licenses_strs)


def new_dpkg_package(
    entry: DpkgDBEntry,
    db_location: Location,
    _resolver: Resolver | None,
    release: Release | None = None,
) -> Package | tuple[Package, Package] | None:
    name = entry.package
    version = entry.version

    if not name or not version:
        return None

    try:
        dpkg = Package(
            name=name,
            version=version,
            licenses=[],
            p_url=package_url(entry, release),
            locations=[db_location],
            type=PackageType.DebPkg,
            metadata=entry,
            found_by=None,
            language=Language.UNKNOWN_LANGUAGE,
        )
        if _resolver is not None:
            # side effects
            merge_file_listing(_resolver, db_location, dpkg)
            add_licenses(_resolver, db_location, dpkg)

        source_dpkg: Package | None = None
        if (entry.source and entry.source != dpkg.name) or (
            entry.source and entry.source_version and entry.source_version != dpkg.version
        ):
            new_entry = deepcopy(entry)
            new_entry.package = entry.source
            new_entry.version = entry.source_version or dpkg.version
            new_entry.source = None
            new_entry.source_version = None
            new_entry.dependencies = None
            new_entry.pre_dependencies = None

            source_dpkg = deepcopy(dpkg)
            source_dpkg.name = new_entry.package
            source_dpkg.version = new_entry.version
            source_dpkg.p_url = package_url(new_entry, release)

        if source_dpkg:
            return dpkg, source_dpkg
        return dpkg
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": db_location.path(),
                },
            },
        )
        return None



import io
import logging
from collections.abc import Generator
from pathlib import Path

from labels.parsers.cataloger.redhat.rpmdb.berkeley.constants import (
    HASH_OF_INDEX_PAGE_TYPE,
    HASH_PAGE_TYPE,
    HASH_UNSORTED_PAGE_TYPE,
    VALID_PAGE_SIZES,
)
from labels.parsers.cataloger.redhat.rpmdb.berkeley.hash_metadata_page import (
    HashMetadataPage,
    parse_hash_metadata_page,
)
from labels.parsers.cataloger.redhat.rpmdb.berkeley.hash_page import (
    hash_page_value_content,
    hash_page_value_indexes,
    parse_hash_page,
)
from labels.parsers.cataloger.redhat.rpmdb.rpmdb_interface import (
    RpmDBInterface,
)
from labels.utils.exceptions import (
    InvalidDBFormatError,
    InvalidMetadataError,
)

LOGGER = logging.getLogger(__name__)


class BerkeleyDB(RpmDBInterface):
    def __init__(self, file_path: str) -> None:
        self.file_path = file_path
        self.file = Path(file_path).open("rb")  # noqa: SIM115
        self.hash_metadata = self._read_metadata()

    def _read_metadata(self) -> HashMetadataPage:
        # Read the first 512 bytes to parse metadata
        self.file.seek(0)
        metadata_buff = self.file.read(512)
        if len(metadata_buff) < 512:
            error_msg = "Failed to read metadata: insufficient data"
            raise InvalidDBFormatError(error_msg)

        # Parse the hash metadata page
        hash_metadata = parse_hash_metadata_page(metadata_buff)

        # Validate the page size
        if hash_metadata.hash_metadata.generic_metadata_page.page_size not in VALID_PAGE_SIZES:
            raise InvalidDBFormatError(
                "Unexpected page size: "
                + str(hash_metadata.hash_metadata.generic_metadata_page.page_size),
            )

        return hash_metadata

    def close(self) -> None:
        self.file.close()

    def read(self) -> Generator[bytes, None, None]:
        page_size = self.hash_metadata.hash_metadata.generic_metadata_page.page_size
        swapped = self.hash_metadata.swapped
        last_page_no = self.hash_metadata.hash_metadata.generic_metadata_page.last_page_no

        self.file.seek(0, io.SEEK_SET)

        for _ in range(last_page_no + 1):
            page_data = self._read_page(page_size)
            if not page_data:
                return

            hash_page_header = parse_hash_page(page_data, swapped=swapped)
            if hash_page_header.page_type not in (
                HASH_UNSORTED_PAGE_TYPE,
                HASH_PAGE_TYPE,
            ):
                continue

            hash_page_indexes = hash_page_value_indexes(
                page_data,
                hash_page_header.num_entries,
                swapped=swapped,
            )

            for hash_page_index in hash_page_indexes:
                if page_data[hash_page_index] != HASH_OF_INDEX_PAGE_TYPE:
                    continue

                current_file_position = self.file.tell()
                try:
                    value_content = hash_page_value_content(
                        self.file,
                        page_data,
                        hash_page_index,
                        page_size,
                        swapped=swapped,
                    )
                    yield value_content
                except ValueError:
                    return
                finally:
                    self.file.seek(current_file_position, io.SEEK_SET)

    def _read_page(self, page_size: int) -> bytes:
        page_data = self.file.read(page_size)
        if len(page_data) != page_size:
            LOGGER.error(
                "Failed to read page",
                extra={"extra": {"location": self.file_path}},
            )
            return b""
        return page_data


def open_berkeley(file_path: str) -> BerkeleyDB:
    try:
        berkeley_db = BerkeleyDB(file_path)
    except InvalidMetadataError as exc:
        LOGGER.warning(
            "Invalid Berkeley database file",
            extra={
                "extra": {
                    "location": file_path,
                },
            },
        )
        raise InvalidDBFormatError from exc
    return berkeley_db



import logging
from typing import NotRequired, TypedDict

from labels.config.cache import dual_cache
from labels.enrichers.api_interface import make_get

LOGGER = logging.getLogger(__name__)


class NPMPackageAuthor(TypedDict):
    email: str
    name: str


class NPMPackageDist(TypedDict):
    integrity: str
    tarball: str


class NPMPackageLicense(TypedDict):
    type: str
    url: str


class NPMPackageVersion(TypedDict):
    dist: NPMPackageDist
    name: str


class NPMPackageTimeUnpublished(TypedDict):
    time: str
    versions: list[str]


NPMPackage = TypedDict(
    "NPMPackage",
    {
        "author": NotRequired[str | NPMPackageAuthor],
        "dist-tags": NotRequired[dict[str, str]],
        "license": str | NPMPackageLicense,
        "name": str,
        "time": dict[str, str],
        "versions": dict[str, NPMPackageVersion],
    },
)


@dual_cache
def get_npm_package(package_name: str) -> NPMPackage | None:
    package: NPMPackage | None = make_get(f"https://registry.npmjs.org/{package_name}", timeout=30)

    if not package:
        return None

    if "unpublished" in package.get("time", {}):
        return None

    return package



import json
import os
from fnmatch import fnmatch
from pathlib import Path
from typing import TextIO, cast

from pydantic import BaseModel

from labels.model.file import Location
from labels.model.indexables import ParsedValue
from labels.parsers.cataloger.python.model import PythonPackage


class ParsedData(BaseModel):
    licenses: str
    license_file: str | list[str]
    license_expresion: str
    license_location: Location | None
    python_package: PythonPackage


def determine_site_package_root_package(path: str) -> str:
    if any(fnmatch(path, pattern) for pattern in ("**/*.egg-info", "**/*.dist-info")):
        return str(Path(path).parent)
    return str(Path(path).parent.parent)


def _split_lines(data: str) -> list[str]:
    # Split the input data into lines
    return data.strip().split("\n")


def _handle_multiline_values(
    line: str,
    multi_line_key: str,
    parsed_data: dict[str, str | list[str]],
) -> str:
    # Append multiline values to the corresponding key
    if multi_line_key and isinstance(parsed_data[multi_line_key], str):
        parsed_data[multi_line_key] += "\n" + line.strip()  # type: ignore
    return multi_line_key


def _update_parsed_data(key: str, value: str, parsed_data: dict[str, str | list[str]]) -> None:
    # Update the parsed data dictionary with new or existing keys
    if key in parsed_data:
        if isinstance(parsed_data[key], list):
            parsed_data[key].append(value)  # type: ignore
        else:
            parsed_data[key] = [parsed_data[key], value]  # type: ignore
    else:
        parsed_data[key] = value


def _process_line(line: str, parsed_data: dict[str, str | list[str]]) -> str:
    # Process each line, updating the parsed data and handling multiline values
    multi_line_key = None
    if ": " in line:
        key, value = line.split(": ", 1)
        _update_parsed_data(key, value, parsed_data)
        if key in ["Description", "Classifier"]:
            multi_line_key = key
    return multi_line_key or ""


def parse_metadata(data: str) -> dict[str, str | list[str]]:
    parsed_data: dict[str, str | list[str]] = {}
    lines = _split_lines(data)
    multi_line_key: str | None = None
    for line in lines:
        if not line:
            break
        if multi_line_key and line.startswith((" ", "\t")):
            multi_line_key = _handle_multiline_values(line, multi_line_key, parsed_data)
        else:
            multi_line_key = _process_line(line, parsed_data)
    return parsed_data


def required_dependencies(
    requires_dis: list[str] | str,
    provides_extra: list[str] | None = None,
) -> list[str]:
    if isinstance(requires_dis, str):
        requires_dis = [requires_dis]
    result: list[str] = []
    provides_extra = provides_extra or []
    for item in requires_dis:
        parts = item.split(";")
        if any(x in parts[-1] for x in provides_extra):
            continue
        result.append(parts[0].strip())
    return result


def _handle_platform(input_value: ParsedValue) -> str | None:
    if isinstance(input_value, list | dict):
        return json.dumps(input_value)
    if isinstance(input_value, str):
        return input_value
    return None


def parse_wheel_or_egg_metadata(path: str, reader: TextIO) -> ParsedData | None:
    metadata_dict: dict[str, str] = cast(dict[str, str], parse_metadata(reader.read()))
    p_data = ParsedData(
        python_package=PythonPackage(
            name=metadata_dict.get("Name", ""),
            version=metadata_dict.get("Version", ""),
            author=metadata_dict.get("Author"),
            author_email=metadata_dict.get("Author-email"),
            platform=_handle_platform(metadata_dict.get("Platform")),
            site_package_root_path=determine_site_package_root_package(path),
            files=None,
            top_level_packages=None,
            direct_url_origin=None,
            dependencies=required_dependencies(
                cast(list[str], metadata_dict["Requires-Dist"]),
                cast(list[str] | None, metadata_dict.get("Provides-Extra")),
            )
            if "Requires-Dist" in metadata_dict
            else None,
        ),
        licenses=metadata_dict.get("License", ""),
        license_expresion=metadata_dict.get("License-Expression", ""),
        license_file=metadata_dict.get("License-File", ""),
        license_location=None,
    )

    if p_data.license_expresion or p_data.licenses:
        p_data.license_file = path
        p_data.license_location = Location(
            access_path=path,
        )
    elif p_data.license_file and isinstance(p_data.license_file, str):
        p_data.license_location = Location(
            access_path=os.path.join(Path(path).parent, p_data.license_file),
        )
    return p_data



import logging
from collections.abc import Iterator
from copy import deepcopy
from typing import cast

from more_itertools import flatten
from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.file import DependencyType, LocationReadCloser
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship, RelationshipType
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.collection.json import parse_json_with_tree_sitter
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)
EMPTY_DICT: IndexedDict[str, ParsedValue] = IndexedDict()


def _get_relationships(
    packages: list[Package],
    dependencies: dict[str, ParsedValue],
) -> list[Relationship]:
    results: list[Relationship] = []
    for package_name, depens in dependencies.items():
        if not isinstance(depens, IndexedDict):
            continue
        deps = (p for p in packages for dep_name in depens if dep_name == p.name)
        if current_package := next(
            (x for x in packages if x.name == package_name),
            None,
        ):
            results.extend(
                Relationship(
                    from_=x,
                    to_=current_package,
                    type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                )
                for x in deps
            )
    return results


def parse_dotnet_package_lock(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_json = cast(
        IndexedDict[str, ParsedValue],
        parse_json_with_tree_sitter(reader.read_closer.read()),
    )
    packages: list[Package] = []
    target_deps: ParsedValue = package_json.get("dependencies", IndexedDict())
    if not isinstance(target_deps, IndexedDict):
        LOGGER.warning("No deps found in package JSON")
        return ([], [])
    dependencies: dict[str, ParsedValue] = {}

    for package_name, package_value in cast(
        Iterator[IndexedDict[str, ParsedValue]],
        flatten(x.items() for x in target_deps.values() if isinstance(x, IndexedDict)),
    ):
        if not isinstance(package_value, IndexedDict):
            continue

        is_transitive: str = package_value.get("type", "") == "Transitive"

        version: str | None = package_value.get("resolved")

        if not package_name or not version:
            continue

        location = deepcopy(reader.location)
        if location.coordinates:
            location.coordinates.line = package_value.position.start.line
            location.dependency_type = (
                DependencyType.TRANSITIVE if is_transitive else DependencyType.DIRECT
            )

        dependencies[package_name] = package_value.get("dependencies", EMPTY_DICT)
        try:
            packages.append(
                Package(
                    name=package_name,
                    version=version,
                    locations=[location],
                    licenses=[],
                    type=PackageType.DotnetPkg,
                    language=Language.DOTNET,
                    metadata=None,
                    p_url=PackageURL(
                        type="nuget",
                        namespace="",
                        name=package_name,
                        version=version,
                        qualifiers={},
                        subpath="",
                    ).to_string(),
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": location.path(),
                    },
                },
            )
            continue

    relationships: list[Relationship] = _get_relationships(packages, dependencies)

    return packages, relationships



from labels.enrichers.elixir.get import get_hex_package
from labels.model.package import HealthMetadata, Package
from labels.utils.licenses.validation import validate_licenses


def complete_package(package: Package) -> Package:
    response = get_hex_package(package.name)
    if not response:
        return package
    package.health_metadata = HealthMetadata(
        latest_version=response["latest_stable_version"],
        latest_version_created_at=next(
            (
                x["inserted_at"]
                for x in response["releases"]
                if x["version"] == response["latest_stable_version"]
            ),
            None,
        ),
    )

    package.licenses = validate_licenses(response["meta"]["licenses"])

    if response["owners"]:
        package.health_metadata.authors = ", ".join([x["username"] for x in response["owners"]])
    return package



import json
import logging
import sqlite3
from typing import Any

from labels.advisories.database import BaseDatabase
from labels.advisories.match_versions import match_vulnerable_versions
from labels.advisories.utils import generate_cpe
from labels.model.advisories import Advisory

LOGGER = logging.getLogger(__name__)


class ImagesDatabase(BaseDatabase):
    def __init__(self) -> None:
        super().__init__(db_name="skims_sca_advisories_for_images.db")


DATABASE = ImagesDatabase()


def fetch_advisory_from_database(
    cursor: sqlite3.Cursor,
    package_manager: str,
    platform_version: str,
    package_name: str,
) -> list[Any]:
    cursor.execute(
        """
        SELECT
            adv_id,
            source,
            vulnerable_version,
            severity_level,
            severity,
            severity_v4,
            epss,
            details,
            percentile,
            cwe_ids,
            cve_finding,
            auto_approve
        FROM advisories
        WHERE package_manager = ? AND platform_version = ? AND package_name = ?;
        """,
        (package_manager, platform_version, package_name),
    )
    return cursor.fetchall()


def create_advisory_from_record(
    result: list[Any],
    package_manager: str,
    package_name: str,
    version: str,
) -> Advisory:
    return Advisory(
        id=result[0],
        urls=[result[1]],
        version_constraint=result[2] or None,
        severity=result[3] or "Low",  # F011 cvss4 severity
        cvss3=result[4],
        cvss4=result[5],
        epss=result[6] or 0.0,
        description=result[7] or None,
        percentile=result[8] or 0.0,
        cpes=[generate_cpe(package_manager, package_name, version)],
        namespace=package_manager,
        cwe_ids=json.loads(result[9]) if result[9] else ["CWE-1395"],
        cve_finding=result[10],
        auto_approve=result[11] or False,
    )


def get_package_advisories(
    package_manager: str,
    package_name: str,
    version: str,
    platform_version: str,
) -> list[Advisory]:
    connection = DATABASE.get_connection()
    cursor = connection.cursor()

    advisories_records = fetch_advisory_from_database(
        cursor,
        package_manager,
        platform_version,
        package_name,
    )

    return [
        create_advisory_from_record(
            result,
            package_manager,
            package_name,
            version,
        )
        for result in advisories_records
    ]


def get_vulnerabilities(
    platform: str,
    product: str,
    version: str,
    platform_version: str | None,
) -> list[Advisory]:
    vulnerabilities = []
    if (
        product
        and version
        and platform_version
        and (
            advisories := get_package_advisories(
                platform,
                product.lower(),
                version,
                platform_version,
            )
        )
    ):
        vulnerabilities = [
            advisor
            for advisor in advisories
            if match_vulnerable_versions(version.lower(), advisor.version_constraint)
        ]

    return vulnerabilities



from labels.enrichers.rust.get import CRATES_ENDPOINT, CargoPackage, Version, get_cargo_package
from labels.enrichers.utils import infer_algorithm
from labels.model.package import Artifact, Digest, HealthMetadata, Package
from labels.utils.licenses.validation import validate_licenses


def _get_artifact(
    current_package: Version | None,
) -> Artifact | None:
    if current_package:
        digest_value = current_package.get("checksum")
        return Artifact(
            url=f"{CRATES_ENDPOINT}{current_package['dl_path']}",
            integrity=Digest(
                algorithm=infer_algorithm(digest_value),
                value=digest_value,
            ),
        )
    return None


def _set_health_metadata(
    package: Package,
    current_package: Version | None,
    cargo_package: CargoPackage,
) -> None:
    crate_info = cargo_package.get("crate", {})
    max_stable_version = crate_info.get("max_stable_version")
    updated_at = crate_info.get("updated_at")
    published_by = cargo_package["versions"][0].get("published_by")
    artifact = _get_artifact(current_package)

    package.health_metadata = HealthMetadata(
        latest_version=max_stable_version,
        latest_version_created_at=updated_at,
        artifact=artifact,
        authors=published_by["name"] if published_by else None,
    )


def complete_package(package: Package) -> Package:
    cargo_package = get_cargo_package(package.name)
    if not cargo_package:
        return package

    versions = cargo_package["versions"]
    current_package = next(
        (version for version in versions if version["num"] == package.version),
        None,
    )

    _set_health_metadata(package, current_package, cargo_package)

    licenses = current_package.get("license") if current_package else None
    if isinstance(licenses, str):
        package.licenses = validate_licenses([licenses])

    return package



import io
import struct
import time
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Callable

from pydantic import BaseModel, Field

from labels.parsers.cataloger.redhat.rpmdb.entry import IndexEntry
from labels.parsers.cataloger.redhat.rpmdb.file_digest import DigestAlgorithm
from labels.parsers.cataloger.redhat.rpmdb.rpmtags import (
    RPM_BIN_TYPE,
    RPM_I18NSTRING_TYPE,
    RPM_INT16_TYPE,
    RPM_INT32_TYPE,
    RPM_STRING_ARRAY_TYPE,
    RPM_STRING_TYPE,
    RPMTAG_ARCH,
    RPMTAG_BASENAMES,
    RPMTAG_DIRINDEXES,
    RPMTAG_DIRNAMES,
    RPMTAG_EPOCH,
    RPMTAG_FILEDIGESTALGO,
    RPMTAG_FILEDIGESTS,
    RPMTAG_FILEFLAGS,
    RPMTAG_FILEGROUPNAME,
    RPMTAG_FILEMODES,
    RPMTAG_FILESIZES,
    RPMTAG_FILEUSERNAME,
    RPMTAG_INSTALLTIME,
    RPMTAG_LICENSE,
    RPMTAG_MODULARITYLABEL,
    RPMTAG_NAME,
    RPMTAG_PGP,
    RPMTAG_PROVIDENAME,
    RPMTAG_RELEASE,
    RPMTAG_REQUIRENAME,
    RPMTAG_SIGMD5,
    RPMTAG_SIZE,
    RPMTAG_SOURCERPM,
    RPMTAG_SUMMARY,
    RPMTAG_VENDOR,
    RPMTAG_VERSION,
)

SIZE_OF_UINT16 = 2
SIZE_OF_INT32 = 4


class PackageInfo(BaseModel):
    epoch: int | None = None
    name: str = ""
    version: str = ""
    release: str = ""
    arch: str = ""
    source_rpm: str = ""
    size: int = 0
    license: str = ""
    vendor: str = ""
    modularity_label: str = ""
    summary: str = ""
    pgp: str = ""
    sig_md5: str = ""
    digest_algorithm: DigestAlgorithm | None = None
    install_time: int = 0
    base_names: list[str] = Field(default_factory=list)
    dir_indexes: list[int] = Field(default_factory=list)
    dir_names: list[str] = Field(default_factory=list)
    file_sizes: list[int] = Field(default_factory=list)
    file_digests: list[str] = Field(default_factory=list)
    file_modes: list[int] = Field(default_factory=list)
    file_flags: list[int] = Field(default_factory=list)
    user_names: list[str] = Field(default_factory=list)
    group_names: list[str] = Field(default_factory=list)
    provides: list[str] = Field(default_factory=list)
    requires: list[str] = Field(default_factory=list)


class PgpSig:
    pub_key_algo: int
    hash_algo: int
    key_id: bytes
    date: int

    def __init__(self, data: bytes) -> None:
        (
            self.date,
            self.key_id,
            self.pub_key_algo,
            self.hash_algo,
        ) = struct.unpack(">3x i 8s BB", data)


class TextSig:
    pub_key_algo: int
    hash_algo: int
    key_id: bytes
    date: int

    def __init__(self, data: bytes) -> None:
        (
            self.pub_key_algo,
            self.hash_algo,
            self.date,
            self.key_id,
        ) = struct.unpack(">2x BB 4x i 4x 8s", data)


class Pgp4Sig:
    pub_key_algo: int
    hash_algo: int
    key_id: bytes
    date: int

    def __init__(self, data: bytes) -> None:
        (
            self.pub_key_algo,
            self.hash_algo,
            self.key_id,
            self.date,
        ) = struct.unpack(">2x BB 17x 8s 2x i", data)


def parse_int32_array(data: bytes, array_size: int) -> list[int]:
    length = array_size // SIZE_OF_INT32

    try:
        # Unpack the byte data into a list of int32 values (big-endian)
        values = struct.unpack(f">{length}i", data[:array_size])
        return list(values)
    except struct.error as exc:
        error_msg = "Failed to read binary data"
        raise ValueError(error_msg) from exc


def parse_int32(data: bytes) -> int:
    try:
        # Unpack a single big-endian int32 value from the data
        (value,) = struct.unpack(">i", data[:4])
        return value
    except struct.error as exc:
        error_msg = "Failed to read binary data"
        raise ValueError(error_msg) from exc


def uint16_array(data: bytes, array_size: int) -> list[int]:
    length = array_size // SIZE_OF_UINT16

    try:
        # Unpack the byte data into a list of uint16 values (big-endian)
        values = struct.unpack(f">{length}H", data[:array_size])
        return list(values)
    except struct.error as exc:
        error_msg = "Failed to read binary data"
        raise ValueError(error_msg) from exc


def parse_string_array(data: bytes) -> list[str]:
    # Decode the bytes to a string and strip trailing null characters
    decoded_str = data.rstrip(b"\x00").decode("utf-8")

    # Split the string by null terminators
    return decoded_str.split("\x00")


def get_nevra(index_entries: list[IndexEntry]) -> PackageInfo:
    package_info = PackageInfo()
    parsing_map: dict[int, Callable[[IndexEntry, PackageInfo], PackageInfo]] = {
        RPMTAG_DIRINDEXES: _parse_rpmtag_dirindex,
        RPMTAG_DIRNAMES: _parse_rpmtag_dirnames,
        RPMTAG_BASENAMES: _parse_rpmtag_basenames,
        RPMTAG_MODULARITYLABEL: _parse_rpmtag_modularitylabel,
        RPMTAG_NAME: _parse_rpmtag_name,
        RPMTAG_EPOCH: _parse_rpmtag_epoch,
        RPMTAG_VERSION: _parse_rpmtag_version,
        RPMTAG_RELEASE: _parse_rpmtag_release,
        RPMTAG_ARCH: _parse_rpmtag_arch,
        RPMTAG_SOURCERPM: _parse_rpmtag_sourcerpm,
        RPMTAG_PROVIDENAME: _parse_rpmtag_providename,
        RPMTAG_REQUIRENAME: _parse_rpmtag_requirename,
        RPMTAG_LICENSE: _parse_rpmtag_license,
        RPMTAG_VENDOR: _parse_rpmtag_vendor,
        RPMTAG_SIZE: _parse_rpmtag_size,
        RPMTAG_FILEDIGESTALGO: _parse_rpmtag_filedigestalgo,
        RPMTAG_FILESIZES: _parse_rpmtag_filesizes,
        RPMTAG_FILEDIGESTS: _parse_rpmtag_filedigests,
        RPMTAG_FILEMODES: _parse_rpmtag_filemodes,
        RPMTAG_FILEFLAGS: _parse_rpmtag_fileflags,
        RPMTAG_FILEUSERNAME: _parse_rpmtag_fileusername,
        RPMTAG_FILEGROUPNAME: _parse_rpmtag_filegroupname,
        RPMTAG_SUMMARY: _parse_rpmtag_summary,
        RPMTAG_INSTALLTIME: _parse_rpmtag_installtime,
        RPMTAG_SIGMD5: _parse_rpmtag_sigmd5,
        RPMTAG_PGP: _parse_rpmtag_pgpsig,
    }
    for entry in index_entries:
        if entry.info.tag not in parsing_map:
            continue
        parse_function = parsing_map[entry.info.tag]
        package_info = parse_function(entry, package_info)
    return package_info


def _parse_rpmtag_dirindex(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_INT32_TYPE:
        error_msg = "Invalid tag dir index"
        raise ValueError(error_msg)
    try:
        package_info.dir_indexes = parse_int32_array(entry.data, entry.length)
    except ValueError as exc:
        error_msg = "Unable to read dir indexes"
        raise ValueError(error_msg) from exc

    return package_info


def _parse_rpmtag_dirnames(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_STRING_ARRAY_TYPE:
        error_msg = "Invalid tag dir names"
        raise ValueError(error_msg)
    try:
        package_info.dir_names = parse_string_array(entry.data)
    except ValueError as exc:
        error_msg = "Unable to read dir names"
        raise ValueError(error_msg) from exc

    return package_info


def _parse_rpmtag_basenames(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_STRING_ARRAY_TYPE:
        error_msg = "Invalid tag base names"
        raise ValueError(error_msg)
    try:
        package_info.base_names = parse_string_array(entry.data)
    except ValueError as exc:
        error_msg = "Unable to read base names"
        raise ValueError(error_msg) from exc
    return package_info


def _parse_rpmtag_modularitylabel(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_STRING_TYPE:
        error_msg = "Invalid tag modularity label"
        raise ValueError(error_msg)
    package_info.modularity_label = entry.data.rstrip(b"\x00").decode("utf-8")
    return package_info


def _parse_rpmtag_name(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_STRING_TYPE:
        error_msg = "Invalid tag name"
        raise ValueError(error_msg)
    package_info.name = entry.data.rstrip(b"\x00").decode("utf-8")
    return package_info


def _parse_rpmtag_epoch(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_INT32_TYPE:
        error_msg = "Invalid tag epoch"
        raise ValueError(error_msg)
    if not entry.data:
        return package_info

    try:
        package_info.epoch = parse_int32(entry.data)
    except ValueError as exc:
        error_msg = "Unable to read epoch"
        raise ValueError(error_msg) from exc
    return package_info


def _parse_rpmtag_version(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_STRING_TYPE:
        error_msg = "Invalid tag version"
        raise ValueError(error_msg)
    package_info.version = entry.data.rstrip(b"\x00").decode("utf-8")
    return package_info


def _parse_rpmtag_release(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_STRING_TYPE:
        error_msg = "Invalid tag release"
        raise ValueError(error_msg)
    package_info.release = entry.data.rstrip(b"\x00").decode("utf-8")
    return package_info


def _parse_rpmtag_arch(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_STRING_TYPE:
        error_msg = "Invalid tag arch"
        raise ValueError(error_msg)
    package_info.arch = entry.data.rstrip(b"\x00").decode("utf-8")
    return package_info


def _parse_rpmtag_sourcerpm(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_STRING_TYPE:
        error_msg = "Invalid tag source rpm"
        raise ValueError(error_msg)
    package_info.source_rpm = entry.data.rstrip(b"\x00").decode("utf-8")

    if package_info.source_rpm == "(none)":
        package_info.source_rpm = ""
    return package_info


def _parse_rpmtag_providename(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_STRING_ARRAY_TYPE:
        error_msg = "Invalid tag provide name"
        raise ValueError(error_msg)
    try:
        package_info.provides = parse_string_array(entry.data)
    except ValueError as exc:
        error_msg = "Unable to read provide name"
        raise ValueError(error_msg) from exc

    return package_info


def _parse_rpmtag_requirename(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_STRING_ARRAY_TYPE:
        error_msg = "Invalid tag require name"
        raise ValueError(error_msg)
    try:
        package_info.requires = parse_string_array(entry.data)
    except ValueError as exc:
        error_msg = "Unable to read require name"
        raise ValueError(error_msg) from exc
    return package_info


def _parse_rpmtag_license(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_STRING_TYPE:
        error_msg = "Invalid tag license"
        raise ValueError(error_msg)

    package_info.license = entry.data.rstrip(b"\x00").decode("utf-8")
    if package_info.license == "(none)":
        package_info.license = ""

    return package_info


def _parse_rpmtag_vendor(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_STRING_TYPE:
        error_msg = "Invalid tag vendor"
        raise ValueError(error_msg)
    package_info.vendor = entry.data.rstrip(b"\x00").decode("utf-8")
    if package_info.vendor == "(none)":
        package_info.vendor = ""
    return package_info


def _parse_rpmtag_size(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    error_msg = "Invalid tag size"
    if entry.info.type != RPM_INT32_TYPE:
        raise ValueError(error_msg)
    try:
        size = parse_int32(entry.data)
    except ValueError as exc:
        raise ValueError(error_msg) from exc
    package_info.size = size
    return package_info


def _parse_rpmtag_filedigestalgo(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_INT32_TYPE:
        error_msg = "Invalid tag file digest algo"
        raise ValueError(error_msg)
    if not entry.data:
        return package_info
    try:
        digest_algorimth = parse_int32(entry.data)
    except ValueError as exc:
        error_msg = "Unable to read file digest algo"
        raise ValueError(error_msg) from exc
    package_info.digest_algorithm = DigestAlgorithm(algorithm=digest_algorimth)
    return package_info


def _parse_rpmtag_filesizes(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_INT32_TYPE:
        error_msg = "Invalid tag file sizes"
        raise ValueError(error_msg)
    if not entry.data:
        return package_info
    try:
        package_info.file_sizes = parse_int32_array(entry.data, entry.length)
    except ValueError as exc:
        error_msg = "Unable to read file sizes"
        raise ValueError(error_msg) from exc
    return package_info


def _parse_rpmtag_filedigests(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_STRING_ARRAY_TYPE:
        error_msg = "Invalid tag file digests"
        raise ValueError(error_msg)

    try:
        package_info.file_digests = parse_string_array(entry.data)
    except ValueError as exc:
        error_msg = "Unable to read file digests"
        raise ValueError(error_msg) from exc

    return package_info


def _parse_rpmtag_filemodes(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_INT16_TYPE:
        error_msg = "Invalid tag file modes"
        raise ValueError(error_msg)
    try:
        package_info.file_modes = uint16_array(entry.data, entry.length)
    except ValueError as exc:
        error_msg = "Unable to read file modules"
        raise ValueError(error_msg) from exc
    return package_info


def _parse_rpmtag_fileflags(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_INT32_TYPE:
        error_msg = "Invalid tag file flags"
        raise ValueError(error_msg)
    try:
        package_info.file_flags = parse_int32_array(entry.data, entry.length)
    except ValueError as exc:
        error_msg = "Unable to read file flags"
        raise ValueError(error_msg) from exc
    return package_info


def _parse_rpmtag_fileusername(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_STRING_ARRAY_TYPE:
        error_msg = "Invalid tag file username"
        raise ValueError(error_msg)
    try:
        package_info.user_names = parse_string_array(entry.data)
    except ValueError as exc:
        error_msg = "Unable to read fileusername"
        raise ValueError(error_msg) from exc
    return package_info


def _parse_rpmtag_filegroupname(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_STRING_ARRAY_TYPE:
        error_msg = "Invalid tag file groupname"
        raise ValueError(error_msg)
    try:
        package_info.group_names = parse_string_array(entry.data)
    except ValueError as exc:
        error_msg = "Unable to read file groupname"
        raise ValueError(error_msg) from exc
    return package_info


def _parse_rpmtag_summary(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type not in (RPM_I18NSTRING_TYPE, RPM_STRING_TYPE):
        error_msg = "Invalid tag summary"
        raise ValueError(error_msg)
    package_info.summary = entry.data.split(b"\x00")[0].decode("utf-8")

    return package_info


def _parse_rpmtag_installtime(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    if entry.info.type != RPM_INT32_TYPE:
        error_msg = "Invalid tag install time"
        raise ValueError(error_msg)

    try:
        package_info.install_time = parse_int32(entry.data)
    except ValueError as exc:
        error_msg = "Unable to read install time"
        raise ValueError from exc
    return package_info


def _parse_rpmtag_sigmd5(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    package_info.sig_md5 = entry.data.hex()
    return package_info


def _parse_rpmtag_pgpsig(entry: IndexEntry, package_info: PackageInfo) -> PackageInfo:
    pub_key_look = {
        0x01: "RSA",
    }
    has_look_up = {
        0x02: "SHA1",
        0x08: "SHA256",
    }

    if entry.info.type != RPM_BIN_TYPE:
        error_msg = "Invalid tag pgp sig"
        raise ValueError(error_msg)
    reader = io.BytesIO(entry.data)
    _, signature_type, version = struct.unpack(">BBB", reader.read(3))

    pub_key_algo = ""
    hash_algo = ""
    package_date = ""
    key_id = b""
    sig: TextSig | PgpSig | Pgp4Sig | None = None
    if signature_type == 0x01:
        if version == 0x1C:
            sig = TextSig(reader.read(struct.calcsize(">2x BB 4x i 4x 8s")))
            pub_key_algo = pub_key_look.get(sig.pub_key_algo, "Unknown")
            hash_algo = has_look_up.get(sig.hash_algo, "Unknown")
            package_date = time.strftime("%a %b %d %H:%M:%S %Y", time.gmtime(sig.date))
            key_id = sig.key_id
        else:
            sig = PgpSig(reader.read(struct.calcsize(">3x i 8s BB")))
            pub_key_algo = pub_key_look.get(sig.pub_key_algo, "Unknown")
            hash_algo = has_look_up.get(sig.hash_algo, "Unknown")
            package_date = time.strftime("%a %b %d %H:%M:%S %Y", time.gmtime(sig.date))
            key_id = sig.key_id
    elif signature_type == 0x02:
        if version == 0x33:
            sig = Pgp4Sig(reader.read(struct.calcsize(">2x BB 17x 8s 2x i")))
            pub_key_algo = pub_key_look.get(sig.pub_key_algo, "Unknown")
            hash_algo = has_look_up.get(sig.hash_algo, "Unknown")
            package_date = time.strftime("%a %b %d %H:%M:%S %Y", time.gmtime(sig.date))
            key_id = sig.key_id
        else:
            sig = PgpSig(reader.read(struct.calcsize(">3x i 8s BB")))
            pub_key_algo = pub_key_look.get(sig.pub_key_algo, "Unknown")
            hash_algo = has_look_up.get(sig.hash_algo, "Unknown")
            package_date = time.strftime("%a %b %d %H:%M:%S %Y", time.gmtime(sig.date))
            key_id = sig.key_id

    package_info.pgp = f"{pub_key_algo}/{hash_algo}, {package_date}, Key ID {key_id.hex()}"

    return package_info



import logging
import re
from copy import deepcopy

from labels.model.file import LocationReadCloser
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.java.package import package_url

LOGGER = logging.getLogger(__name__)

GRADLE_DIST = re.compile("^distributionUrl=.+gradle-(?P<gradle_version>[^-]+)-.+")


def parse_gradle_properties(
    _resolver: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages = []
    for line_no, raw_line in enumerate(reader.read_closer.readlines(), start=1):
        line = raw_line.strip()
        if (match := GRADLE_DIST.match(line)) and (version := match.group("gradle_version")):
            location = deepcopy(reader.location)
            if location.coordinates:
                location.coordinates.line = line_no

            packages.append(
                Package(
                    name="gradle",
                    version=version,
                    locations=[location],
                    language=Language.JAVA,
                    type=PackageType.JavaPkg,
                    p_url=package_url("gradle", version),
                    licenses=[],
                ),
            )

    return packages, []



from typing import NotRequired, TypedDict, cast

import requests

from labels.config.cache import dual_cache


class PyPIInfo(TypedDict):
    author: str
    author_email: str
    bugtrack_url: NotRequired[str]
    classifiers: list[str]
    description: str
    description_content_type: NotRequired[str]
    docs_url: NotRequired[str]
    download_url: NotRequired[str]
    downloads: dict[str, int]
    dynamic: NotRequired[str]
    home_page: NotRequired[str]
    keywords: NotRequired[str]
    license: str
    license_expression: NotRequired[str]
    license_files: NotRequired[str]
    maintainer: NotRequired[str]
    maintainer_email: NotRequired[str]
    name: str
    package_url: str
    platform: NotRequired[str]
    project_url: str
    project_urls: NotRequired[dict[str, str]]
    provides_extra: NotRequired[list[str]]
    release_url: str
    requires_dist: list[str]
    requires_python: str
    summary: str
    version: str
    yanked: bool
    yanked_reason: NotRequired[str]


class PyPIUrl(TypedDict):
    comment_text: str
    digests: dict[str, str]
    downloads: int
    filename: str
    has_sig: bool
    md5_digest: str
    packagetype: str
    python_version: str
    requires_python: str
    size: int
    upload_time: str
    upload_time_iso_8601: str
    url: str
    yanked: bool
    yanked_reason: NotRequired[str]


class PyPIResponse(TypedDict):
    info: PyPIInfo
    last_serial: int
    releases: dict[str, list[dict[str, object]]]
    urls: list[PyPIUrl]


@dual_cache
def get_pypi_package(package_name: str, version: str | None = None) -> PyPIResponse | None:
    url = f"https://pypi.org/pypi/{package_name}{'/' + version if version else ''}/json"
    response = requests.get(url, timeout=30)
    if response.status_code == 200:
        return cast(PyPIResponse, response.json())
    return None



import logging
import re
from contextlib import suppress

from bs4 import BeautifulSoup
from pydantic import ValidationError

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.javascript.package import package_url
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)

SCRIPT_DEP = re.compile(
    r"(?P<name>[^\s\/]*)(?P<separator>[-@\/])"
    r"(?P<version>(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*))",
)


def _get_location(location: Location, sourceline: int) -> Location:
    if location.coordinates:
        return location.model_copy(
            update={"coordinates": location.coordinates.model_copy(update={"line": sourceline})},
        )
    return location


def parse_html_scripts(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    html = None
    with suppress(UnicodeError):
        try:
            html = BeautifulSoup(reader.read_closer, features="html.parser")
        except AssertionError:
            return [], []

    if not html:
        return [], []

    packages = []
    for script in html("script"):
        src_attribute = script.attrs.get("src")
        if not (src_attribute and src_attribute.endswith(".js")):
            continue

        matched = SCRIPT_DEP.search(src_attribute)

        if not matched:
            continue

        name = matched.group("name")
        version = matched.group("version")

        if not name or not version:
            continue

        location = _get_location(reader.location, script.sourceline)

        try:
            packages.append(
                Package(
                    name=name,
                    version=version,
                    licenses=[],
                    locations=[location],
                    language=Language.JAVASCRIPT,
                    type=PackageType.NpmPkg,
                    p_url=package_url(name, version),
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": location.path(),
                    },
                },
            )
            continue

    return packages, []



import base64

from labels.enrichers.dotnet.get import NugetCatalogEntry, get_nuget_package
from labels.enrichers.utils import infer_algorithm
from labels.model.package import Artifact, Digest, HealthMetadata, Package
from labels.utils.licenses.validation import validate_licenses


def _get_artifact(current_package: NugetCatalogEntry | None) -> Artifact | None:
    if current_package:
        # {LOWER_ID}/{LOWER_VERSION}/{LOWER_ID}.{LOWER_VERSION}.nupkg
        lower_id: str | None = current_package["id"].lower()
        lower_version: str | None = current_package["version"].lower()

        digest_value: str | None = current_package.get("packageHash") or None
        algorithm: str | None = current_package.get("packageHashAlgorithm") or None

        if algorithm:
            algorithm = algorithm.lower()
            if algorithm == "sha512" and digest_value:
                digest_value = base64.b64decode(digest_value).hex()

        return Artifact(
            url=(
                f"https://api.nuget.org/v3-flatcontainer/{lower_id}"
                f"/{lower_version}/{lower_id}.{lower_version}.nupkg"
            ),
            integrity=Digest(
                algorithm=infer_algorithm(digest_value),
                value=digest_value,
            ),
        )
    return None


def _set_health_metadata(
    package: Package,
    nuget_package: NugetCatalogEntry,
    current_package: NugetCatalogEntry | None,
) -> None:
    package.health_metadata = HealthMetadata(
        latest_version=nuget_package.get("version"),
        latest_version_created_at=nuget_package.get("published"),
        authors=nuget_package.get("authors"),
        artifact=_get_artifact(current_package) if current_package else None,
    )


def complete_package(package: Package) -> Package:
    current_package = get_nuget_package(package.name, package.version)
    nuget_package = get_nuget_package(package.name)

    if not nuget_package:
        return package

    _set_health_metadata(package, nuget_package, current_package)

    nuget_licenses = nuget_package["licenseExpression"]
    package.licenses = validate_licenses([nuget_licenses]) if nuget_licenses else []

    return package



import re
from copy import (
    deepcopy,
)

from packageurl import (
    PackageURL,
)

from labels.model.file import (
    DependencyType,
    LocationReadCloser,
)
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import (
    Relationship,
)
from labels.model.resolver import (
    Resolver,
)
from labels.parsers.cataloger.generic.parser import (
    Environment,
)

QUOTE = r'["\']'
NL = r"(\n?\s*)?"
TEXT = r'[^"\']+'
RE_SBT: re.Pattern[str] = re.compile(
    r"^[^%]*"
    rf"{NL}{QUOTE}(?P<group>{TEXT}){QUOTE}{NL}%"
    rf"{NL}{QUOTE}(?P<name>{TEXT}){QUOTE}{NL}%"
    rf"{NL}{QUOTE}(?P<version>{TEXT}){QUOTE}{NL}"
    r".*$",
)

VERSION_VAR_RE: re.Pattern[str] = re.compile(
    rf"^\s*val\s+(?P<var_name>\w+)(?:\s*:\s*\w+)?\s*=\s*{QUOTE}(?P<value>{TEXT}){QUOTE}\s*$",
)

VERSION_REF_RE: re.Pattern[str] = re.compile(
    rf"{QUOTE}(?P<group>{TEXT}){QUOTE}\s*%%?\s*{QUOTE}(?P<name>{TEXT}){QUOTE}\s*%\s*(?P<version_var>\w+)",
)


def build_stb(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages = []
    version_vars = {}
    content = reader.read_closer.read()

    for line in content.splitlines():
        if match := VERSION_VAR_RE.match(line):
            version_vars[match.group("var_name")] = match.group("value")

    for line_no, line in enumerate(content.splitlines(), start=1):
        if match := RE_SBT.match(line):
            product = str(match.group("group")) + ":" + match.group("name")
            version = match.group("version")
        elif match := VERSION_REF_RE.search(line):
            product = str(match.group("group")) + ":" + match.group("name")
            version_var = match.group("version_var")
            version = version_vars.get(version_var, "unknown")
        else:
            continue

        location = deepcopy(reader.location)
        if location.coordinates:
            location.coordinates.line = line_no
            location.dependency_type = DependencyType.DIRECT
        packages.append(
            Package(
                name=product,
                version=version,
                type=PackageType.JavaPkg,
                locations=[location],
                p_url=PackageURL(
                    type="maven",
                    namespace=str(match.group("group")),
                    name=str(match.group("name")),
                    version=version,
                    qualifiers=None,
                    subpath="",
                ).to_string(),
                metadata=None,
                language=Language.JAVA,
                licenses=[],
                is_dev=False,
            ),
        )
    return packages, []
